# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

__version_info__ = ("0", "1", "1")
__version__ = ".".join(__version_info__)



#!/usr/bin/python

import pandas as pd
import numpy as np
import re
import sys
import os
import argparse
import matplotlib
from matplotlib import pyplot as plt
from matplotlib.projections.polar import PolarAxes
from matplotlib.projections import register_projection

"""
Read the stats file produced by the OpenMP runtime
and produce a processed summary

The radar_factory original code was taken from
matplotlib.org/examples/api/radar_chart.html
We added support to handle negative values for radar charts
"""

def radar_factory(num_vars, frame='circle'):
    """Create a radar chart with num_vars axes."""
    # calculate evenly-spaced axis angles
    theta = 2*np.pi * np.linspace(0, 1-1./num_vars, num_vars)
    # rotate theta such that the first axis is at the top
    #theta += np.pi/2

    def draw_poly_frame(self, x0, y0, r):
        # TODO: use transforms to convert (x, y) to (r, theta)
        verts = [(r*np.cos(t) + x0, r*np.sin(t) + y0) for t in theta]
        return plt.Polygon(verts, closed=True, edgecolor='k')

    def draw_circle_frame(self, x0, y0, r):
        return plt.Circle((x0, y0), r)

    frame_dict = {'polygon': draw_poly_frame, 'circle': draw_circle_frame}
    if frame not in frame_dict:
        raise ValueError, 'unknown value for `frame`: %s' % frame

    class RadarAxes(PolarAxes):
        """
        Class for creating a radar chart (a.k.a. a spider or star chart)

        http://en.wikipedia.org/wiki/Radar_chart
        """
        name = 'radar'
        # use 1 line segment to connect specified points
        RESOLUTION = 1
        # define draw_frame method
        draw_frame = frame_dict[frame]

        def fill(self, *args, **kwargs):
            """Override fill so that line is closed by default"""
            closed = kwargs.pop('closed', True)
            return super(RadarAxes, self).fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super(RadarAxes, self).plot(*args, **kwargs)
            #for line in lines:
            #    self._close_line(line)

        def set_varlabels(self, labels):
            self.set_thetagrids(theta * 180/np.pi, labels,fontsize=14)

        def _gen_axes_patch(self):
            x0, y0 = (0.5, 0.5)
            r = 0.5
            return self.draw_frame(x0, y0, r)

    register_projection(RadarAxes)
    return theta

# Code to read the raw stats
def extractSI(s):
    """Convert a measurement with a range suffix into a suitably scaled value"""
    du     = s.split()
    num    = float(du[0])
    units  = du[1] if len(du) == 2 else ' '
    # http://physics.nist.gov/cuu/Units/prefixes.html
    factor = {'Y':  1e24,
              'Z':  1e21,
              'E':  1e18,
              'P':  1e15,
              'T':  1e12,
              'G':  1e9,
              'M':  1e6,
              'k':  1e3,
              ' ':  1  ,
              'm': -1e3, # Yes, I do mean that, see below for the explanation.
              'u': -1e6,
              'n': -1e9,
              'p': -1e12,
              'f': -1e15,
              'a': -1e18,
              'z': -1e21,
              'y': -1e24}[units[0]]
    # Minor trickery here is an attempt to preserve accuracy by using a single
    # divide, rather than  multiplying by 1/x, which introduces two roundings
    # since 1/10 is not representable perfectly in IEEE floating point. (Not
    # that this really matters, other than for cleanliness, since we're likely
    # reading numbers with at most five decimal digits of precision).
    return  num*factor if factor > 0 else num/-factor

def readData(f):
    line = f.readline()
    fieldnames = [x.strip() for x in line.split(',')]
    line = f.readline().strip()
    data = []
    while line != "":
        if line[0] != '#':
            fields = line.split(',')
            data.append ((fields[0].strip(), [extractSI(v) for v in fields[1:]]))
        line = f.readline().strip()
    # Man, working out this next incantation out was non-trivial!
    # They really want you to be snarfing data in csv or some other
    # format they understand!
    res = pd.DataFrame.from_items(data, columns=fieldnames[1:], orient='index')
    return res

def readTimers(f):
    """Skip lines with leading #"""
    line = f.readline()
    while line[0] == '#':
        line = f.readline()
    line = line.strip()
    if line == "Statistics on exit\n" or "Aggregate for all threads\n":
        line = f.readline()
    return readData(f)

def readCounters(f):
    """This can be just the same!"""
    return readData(f)

def readFile(fname):
    """Read the statistics from the file. Return a dict with keys "timers", "counters" """
    res = {}
    try:
        with open(fname) as f:
            res["timers"]   = readTimers(f)
            res["counters"] = readCounters(f)
            return res
    except (OSError, IOError):
        print "Cannot open " + fname
        return None

def usefulValues(l):
    """I.e. values which are neither null nor zero"""
    return [p and q for (p,q) in zip (pd.notnull(l), l != 0.0)]

def uselessValues(l):
    """I.e. values which are null or zero"""
    return [not p for p in usefulValues(l)]

interestingStats = ("counters", "timers")
statProperties   = {"counters" : ("Count", "Counter Statistics"),
                    "timers"   : ("Time (ticks)", "Timer Statistics")
                   }

def drawChart(data, kind, filebase):
    """Draw a summary bar chart for the requested data frame into the specified file"""
    data["Mean"].plot(kind="bar", logy=True, grid=True, colormap="GnBu",
                      yerr=data["SD"], ecolor="black")
    plt.xlabel("OMP Constructs")
    plt.ylabel(statProperties[kind][0])
    plt.title (statProperties[kind][1])
    plt.tight_layout()
    plt.savefig(filebase+"_"+kind)

def normalizeValues(data, countField, factor):
    """Normalize values into a rate by dividing them all by the given factor"""
    data[[k for k in data.keys() if k != countField]] /= factor


def setRadarFigure(titles):
    """Set the attributes for the radar plots"""
    fig = plt.figure(figsize=(9,9))
    rect = [0.1, 0.1, 0.8, 0.8]
    labels = [0.2, 0.4, 0.6, 0.8, 1, 2, 3, 4, 5, 10]
    matplotlib.rcParams.update({'font.size':13})
    theta = radar_factory(len(titles))
    ax = fig.add_axes(rect, projection='radar')
    ax.set_rgrids(labels)
    ax.set_varlabels(titles)
    ax.text(theta[2], 1, "Linear->Log", horizontalalignment='center', color='green', fontsize=18)
    return {'ax':ax, 'theta':theta}


def drawRadarChart(data, kind, filebase, params, color):
    """Draw the radar plots"""
    tmp_lin = data * 0
    tmp_log = data * 0
    for key in data.keys():
        if data[key] >= 1:
           tmp_log[key] = np.log10(data[key])
        else:
           tmp_lin[key] = (data[key])
    params['ax'].plot(params['theta'], tmp_log, color='b', label=filebase+"_"+kind+"_log")
    params['ax'].plot(params['theta'], tmp_lin, color='r', label=filebase+"_"+kind+"_linear")
    params['ax'].legend(loc='best', bbox_to_anchor=(1.4,1.2))
    params['ax'].set_rlim((0, np.ceil(max(tmp_log))))

def multiAppBarChartSettings(ax, plt, index, width, n, tmp, s):
    ax.set_yscale('log')
    ax.legend()
    ax.set_xticks(index + width * n / 2)
    ax.set_xticklabels(tmp[s]['Total'].keys(), rotation=50, horizontalalignment='right')
    plt.xlabel("OMP Constructs")
    plt.ylabel(statProperties[s][0])
    plt.title(statProperties[s][1])
    plt.tight_layout()

def derivedTimerStats(data):
    stats = {}
    for key in data.keys():
        if key == 'OMP_worker_thread_life':
            totalRuntime = data['OMP_worker_thread_life']
        elif key in ('FOR_static_iterations', 'OMP_PARALLEL_args',
                     'OMP_set_numthreads', 'FOR_dynamic_iterations'):
            break
        else:
            stats[key] = 100 * data[key] / totalRuntime
    return stats

def compPie(data):
    compKeys = {}
    nonCompKeys = {}
    for key in data.keys():
        if key in ('OMP_critical', 'OMP_single', 'OMP_serial',
                   'OMP_parallel', 'OMP_master', 'OMP_task_immediate',
                   'OMP_task_taskwait', 'OMP_task_taskyield', 'OMP_task_taskgroup',
                   'OMP_task_join_bar', 'OMP_task_plain_bar', 'OMP_task_taskyield'):
            compKeys[key] = data[key]
        else:
            nonCompKeys[key] = data[key]
    print "comp keys:", compKeys, "\n\n non comp keys:", nonCompKeys
    return [compKeys, nonCompKeys]

def drawMainPie(data, filebase, colors):
    sizes = [sum(data[0].values()), sum(data[1].values())]
    explode = [0,0]
    labels = ["Compute - " + "%.2f" % sizes[0], "Non Compute - " + "%.2f" % sizes[1]]
    patches = plt.pie(sizes, explode, colors=colors, startangle=90)
    plt.title("Time Division")
    plt.axis('equal')
    plt.legend(patches[0], labels, loc='best', bbox_to_anchor=(-0.1,1), fontsize=16)
    plt.savefig(filebase+"_main_pie", bbox_inches='tight')

def drawSubPie(data, tag, filebase, colors):
    explode = []
    labels = data.keys()
    sizes = data.values()
    total = sum(sizes)
    percent = []
    for i in range(len(sizes)):
        explode.append(0)
        percent.append(100 * sizes[i] / total)
        labels[i] = labels[i] + " - %.2f" % percent[i]
    patches = plt.pie(sizes, explode=explode, colors=colors, startangle=90)
    plt.title(tag+"(Percentage of Total:"+" %.2f" % (sum(data.values()))+")")
    plt.tight_layout()
    plt.axis('equal')
    plt.legend(patches[0], labels, loc='best', bbox_to_anchor=(-0.1,1), fontsize=16)
    plt.savefig(filebase+"_"+tag, bbox_inches='tight')

def main():
    parser = argparse.ArgumentParser(description='''This script takes a list
        of files containing each of which contain output from a stats-gathering
        enabled OpenMP runtime library.  Each stats file is read, parsed, and
        used to produce a summary of the statistics''')
    parser.add_argument('files', nargs='+',
        help='files to parse which contain stats-gathering output')
    command_args = parser.parse_args()
    colors = ['orange', 'b', 'r', 'yellowgreen', 'lightsage', 'lightpink',
              'green', 'purple', 'yellow', 'cyan', 'mediumturquoise',
              'olive']
    stats = {}
    matplotlib.rcParams.update({'font.size':22})
    for s in interestingStats:
        fig, ax = plt.subplots()
        width = 0.45
        n = 0
        index = 0

        for f in command_args.files:
            filebase = os.path.splitext(f)[0]
            tmp = readFile(f)
            data = tmp[s]['Total']
            """preventing repetition by removing rows similar to Total_OMP_work
                as Total_OMP_work['Total'] is same as OMP_work['Total']"""
            if s == 'counters':
                elapsedTime = tmp["timers"]["Mean"]["OMP_worker_thread_life"]
                normalizeValues(tmp["counters"], "SampleCount",
                    elapsedTime / 1.e9)
                """Plotting radar charts"""
                params = setRadarFigure(data.keys())
                chartType = "radar"
                drawRadarChart(data, s, filebase, params, colors[n])
                """radar Charts finish here"""
                plt.savefig(filebase+"_"+s+"_"+chartType, bbox_inches='tight')
            elif s == 'timers':
                print "overheads in "+filebase
                numThreads = tmp[s]['SampleCount']['Total_OMP_parallel']
                for key in data.keys():
                    if key[0:5] == 'Total':
                        del data[key]
                stats[filebase] = derivedTimerStats(data)
                dataSubSet = compPie(stats[filebase])
                drawMainPie(dataSubSet, filebase, colors)
                plt.figure(0)
                drawSubPie(dataSubSet[0], "Computational Time", filebase, colors)
                plt.figure(1)
                drawSubPie(dataSubSet[1], "Non Computational Time", filebase, colors)
                with open('derivedStats_{}.csv'.format(filebase), 'w') as f:
                    f.write('================={}====================\n'.format(filebase))
                    f.write(pd.DataFrame(stats[filebase].items()).to_csv()+'\n')
            n += 1
    plt.close()

if __name__ == "__main__":
    main()



# MIT License
#
# Copyright (c) 2020, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Imports common utilities
"""

from __future__ import absolute_import

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

import sys
import os
from os.path import dirname
from os.path import basename
from os.path import join

__all__ = [
    "file",
    "func",
    "line",
    "frame",
    "is_generator",
    "is_coroutine",
    "execute_",
    "PY3",
    "PY35",
    "FILE",
    "FUNC",
    "LINE",
    "FRAME",
    "popen",
    "dart_measurement",
    "dart_measurement_file",
    "write_ctest_notes",
    "get_bin_script",
]

CO_GENERATOR = 0x0020
# Python 2/3 compatibility utils
PY3 = sys.version_info[0] == 3
PY35 = PY3 and sys.version_info[1] >= 5
PY37 = PY3 and sys.version_info[1] >= 7


def frame(back=2):
    """Returns a frame"""
    return sys._getframe(back)


def file(back=2, only_basename=True, use_dirname=False, noquotes=True):
    """
    Returns the file name
    """

    def get_fcode(back):
        fname = "<module>"
        try:
            fname = sys._getframe(back).f_code.co_filename
        except Exception as e:
            print(e)
            fname = "<module>"
        return fname

    result = None
    if only_basename is True:
        if use_dirname is True:
            result = "{}".format(
                join(
                    basename(dirname(get_fcode(back))),
                    basename(get_fcode(back)),
                )
            )
        else:
            result = "{}".format(basename(get_fcode(back)))
    else:
        result = "{}".format(get_fcode(back))

    if noquotes is False:
        result = "'{}'".format(result)

    return result


def func(back=2):
    """
    Returns the function name
    """
    return "{}".format(sys._getframe(back).f_code.co_name)


def line(back=1):
    """
    Returns the line number
    """
    return int(sys._getframe(back).f_lineno)


FRAME = frame
FILE = file
FUNC = func
LINE = line


def is_generator(f):
    """Return True if a function is a generator."""
    isgen = (f.__code__.co_flags & CO_GENERATOR) != 0
    return isgen


# exec (from https://bitbucket.org/gutworth/six/):
if PY3:
    import builtins

    execute_ = getattr(builtins, "exec")
    del builtins
else:

    def execute_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")


if PY35:
    import inspect

    def is_coroutine(f):
        return inspect.iscoroutinefunction(f)


else:

    def is_coroutine(f):
        return False


def popen(cmd, outf=None, keep_going=True, timeout=None, shell=False):
    """Execute a command and write to a file or get the output"""

    def handle_error(ret, cmd, keep_going):
        """Handle error messaging for executed command"""
        err_msg = "Error executing: '{}'".format(" ".join(cmd))
        if ret != 0 and not keep_going:
            raise RuntimeError(err_msg)
        elif ret != 0 and keep_going:
            barrier = "=" * 80
            err_msg = (
                "\n\n"
                + barrier
                + "\n\n    ERROR: "
                + err_msg
                + "\n\n"
                + barrier
                + "\n\n"
            )
            sys.stderr.write(err_msg)
            sys.stderr.flush()

    import subprocess as sp

    p = None
    outs = None
    errs = None
    retc = 0
    if shell:
        if PY37:
            p = sp.run(
                " ".join(cmd),
                shell=True,
                check=False,
                capture_output=True,
            )
        else:
            p = sp.run(
                " ".join(cmd),
                shell=True,
                check=False,
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
    else:
        p = sp.Popen(cmd)
        try:
            if timeout is not None:
                outs, errs = p.communicate(timeout=timeout)
            else:
                outs, errs = p.communicate()
        except sp.TimeoutExpired:
            p.kill()

    retc = p.returncode
    if retc != 0:
        if errs is not None:
            print("{}".format(errs.decode("utf-8")))
        elif p.stderr is not None:
            print("{}".format(p.stderr.decode("utf-8")))

    handle_error(retc, cmd, keep_going)

    if outf is not None:
        with open(outf, "w") as f:
            if outs is not None:
                f.write(outs)
            elif p.stdout is not None:
                f.write(p.stdout)

    if outs is None and p.stdout is not None:
        outs = "{}".format(p.stdout.decode("utf-8"))

    if errs is None and p.stderr is not None:
        errs = "{}".format(p.stderr.decode("utf-8"))

    return (retc, outs, errs)


def which(exe):
    """Portable implementation of which"""
    try:
        import shutil

        return shutil.which(exe)
    except (ImportError, AttributeError):
        try:
            from distutils.spawn import find_executable

            return find_executable(exe)
        except (ImportError):
            import os

            _path = os.getenv("PATH")
            for p in _path.split(os.path.pathsep):
                p = os.path.join(p, exe)
                if os.path.exists(p) and os.access(p, os.X_OK):
                    return p

    return None


def dart_measurement(name, value):
    """Prints out an XML tag which gets detected by CTest and recorded by CDash"""
    print(
        f'<DartMeasurement name="{name}" type="numeric/double">{value}</DartMeasurement>'
    )


def dart_measurement_file(name, path, format="png", type="image"):
    """Prints out an XML tag which gets detected by CTest and uploaded to CDash"""
    _path = os.path.abspath(path)
    tag = "DartMeasurementFile"
    print(f'<{tag} name="{name}" type="{type}/{format}">{_path}</{tag}>')


def write_ctest_notes(fname, path=None, mode="a"):
    """Writes or appends to a CTestNotes.txt file"""

    if path is None:
        path = os.path.abspath(os.path.dirname(fname))
    if not os.path.exists(path):
        os.makedirs(path)
    fname = os.path.abspath(fname)
    with open(os.path.join(path, "CTestNotes.txt"), "a") as ofs:
        ofs.write(f'\nlist(APPEND CTEST_NOTES_FILES "{fname}")\n')


def get_bin_script(fname):
    """Returns the path to a script in the internal bin folder"""
    _this_dir = os.path.dirname(__file__)
    _fpath = os.path.join(_this_dir, "bin", fname)
    if os.path.exists(_fpath):
        return _fpath
    import glob

    for itr in glob.glob(f"{_fpath}*"):
        if os.path.exists(itr):
            return itr
    return None



# Runtime report tests

import json
import os
import unittest

import calipertest as cat

class CaliperRuntimeReportTest(unittest.TestCase):
    """ Runtime report controller """
    def test_runtime_report_default(self):
        target_cmd = [ './ci_test_macros', '10', 'runtime-report,output=stdout' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Path',
            'main',
            '  mainloop',
            '    foo',
            '      fooloop'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_runtime_report_nompi(self):
        target_cmd = [ './ci_test_macros', '10', 'runtime-report,aggregate_across_ranks=false,output=stdout' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Time (E) Time (I) Time % (E) Time % (I)',
            'main',
            '  mainloop',
            '    foo',
            '      fooloop'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

if __name__ == "__main__":
    unittest.main()



# File and directory creation test cases

import os
import unittest

import calipertest as cat

class CaliperFileIOTest(unittest.TestCase):
    """ Caliper file and directory creation test case """

    def test_createdir(self):
        target_cmd = [ './ci_test_macros', '0', 'hatchet-region-profile,use.mpi=false,output.format=json,output=foo/bar/test.json' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        if os.path.exists('foo/bar/test.json'):
            os.remove('foo/bar/test.json')
        if os.path.exists('foo/bar'):
            os.removedirs('foo/bar')

        self.assertFalse(os.path.exists('foo/bar'))

        cat.run_test(target_cmd, caliper_config)

        self.assertTrue(os.path.isdir('foo/bar'))
        self.assertTrue(os.path.isfile('foo/bar/test.json'))

        if os.path.exists('foo/bar/test.json'):
            os.remove('foo/bar/test.json')
        if os.path.exists('foo/bar'):
            os.removedirs('foo/bar')


if __name__ == "__main__":
    unittest.main()



#!/bin/python
""" Script for deriving topdown metrics from a Json file produced by Caliper. """

from __future__ import print_function
import sys
import pandas as pd

pd.set_option('display.expand_frame_repr', False)

METRICS = [
    'retiring',
    'bad_speculation',
    'frontend_bound',
    'backend_bound',
    'branch_mispredict',
    'machine_clear',
    'frontend_latency',
    'frontend_bandwidth',
    'memory_bound',
    'core_bound',
    'mem_bound',
    'l1_bound',
    'l2_bound',
    'l3_bound',
    'uncore_bound',
]


def eprint(*args, **kwargs):
    """ Print to stderr """

    print(*args, file=sys.stderr, **kwargs)


def derive_topdown_ivb(dfm):
    """ Perform topdown metric calculations for ivybridge architecture """

    dfm['TEMPORARY_clocks'] = dfm['libpfm.counter.CPU_CLK_UNHALTED.THREAD_P']
    dfm['TEMPORARY_slots'] = 4*dfm['TEMPORARY_clocks']

    # Level 1 - Not Stalled
    dfm['retiring'] = (dfm['libpfm.counter.UOPS_RETIRED.RETIRE_SLOTS']
                       / dfm['TEMPORARY_slots'])
    dfm['bad_speculation'] = ((dfm['libpfm.counter.UOPS_ISSUED.ANY']
                               - dfm['libpfm.counter.UOPS_RETIRED.RETIRE_SLOTS']
                               + 4*dfm['libpfm.counter.INT_MISC.RECOVERY_CYCLES'])
                              / dfm['TEMPORARY_slots'])

    # Level 1 - Stalled
    dfm['frontend_bound'] = (dfm['libpfm.counter.IDQ_UOPS_NOT_DELIVERED.CORE']
                             / dfm['TEMPORARY_slots'])
    dfm['backend_bound'] = (1 - (dfm['frontend_bound']
                                 + dfm['bad_speculation']
                                 + dfm['retiring']))

    # Level 2 - Retiring
    # TODO: implement if possible

    # Level 2 - Bad speculation
    dfm['branch_mispredict'] = (dfm['libpfm.counter.BR_MISP_RETIRED.ALL_BRANCHES']
                                / (dfm['libpfm.counter.BR_MISP_RETIRED.ALL_BRANCHES']
                                   + dfm['libpfm.counter.MACHINE_CLEARS.COUNT']))
    dfm['machine_clear'] = (1 - dfm['branch_mispredict'])  # FIXME: is this correct?

    # Level 2 - Frontend Bound
    dfm['frontend_latency'] = (dfm['libpfm.counter.IDQ_UOPS_NOT_DELIVERED.CORE'].clip(lower=4)
                                     / dfm['TEMPORARY_clocks'])
    dfm['frontend_bandwidth'] = (1 - dfm['frontend_latency'])  # FIXME: is this correct?

    # Level 2 - Backend Bound
    dfm['memory_bound'] = (dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_LDM_PENDING']
                           / dfm['TEMPORARY_clocks'])
    dfm['TEMPORARY_be_bound_at_exe'] = ((dfm['libpfm.counter.CYCLE_ACTIVITY.CYCLES_NO_EXECUTE']
                                         + dfm['libpfm.counter.UOPS_EXECUTED.CORE_CYCLES_GE_1']
                                         - dfm['libpfm.counter.UOPS_EXECUTED.CORE_CYCLES_GE_2'])
                                        / dfm['TEMPORARY_clocks'])
    dfm['core_bound'] = (dfm['TEMPORARY_be_bound_at_exe']
                         - dfm['memory_bound'])

    # Level 3 - Memory bound
    dfm['TEMPORARY_l3_hit_fraction'] = (dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_HIT'] /
                                        (dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_HIT']
                                         + 7*dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_MISS']))
    dfm['TEMPORARY_l3_miss_fraction'] = (7*dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_MISS']
                                         / (dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_HIT']
                                            + 7*dfm['libpfm.counter.MEM_LOAD_UOPS_RETIRED.L3_MISS']))
    dfm['mem_bound'] = (dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L2_PENDING']
                        * dfm['TEMPORARY_l3_miss_fraction']
                        / dfm['TEMPORARY_clocks'])
    dfm['l1_bound'] = ((dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_LDM_PENDING']
                        - dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L1D_PENDING'])
                       / dfm['TEMPORARY_clocks'])
    dfm['l2_bound'] = ((dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L1D_PENDING']
                        - dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L2_PENDING'])
                       / dfm['TEMPORARY_clocks'])
    dfm['l3_bound'] = (dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L2_PENDING']
                       * dfm['TEMPORARY_l3_hit_fraction']
                       / dfm['TEMPORARY_clocks'])
    dfm['uncore_bound'] = (dfm['libpfm.counter.CYCLE_ACTIVITY.STALLS_L2_PENDING']
                           / dfm['TEMPORARY_clocks'])

    for column in dfm.columns:
        if 'TEMPORARY' in column:
            del dfm[column]
        elif 'libpfm.counter' in column:
            del dfm[column]

    return dfm


def derive_topdown(dfm, arch):
    """ Determine topdown function to use, use it, then clean up the dataframe """

    if arch == 'ivybridge':
        dfm = derive_topdown_ivb(dfm)
    else:
        eprint("Error, unsupported architecture " + arch)
        return dfm

    return dfm


def max_column(row, columns):
    """ Returns the column name for the column with the largest value in row """

    return max([(column, row[column]) for column in columns],
               key=lambda t: t[1])[0]


def percentage_string(val):
    """ Returns a percentage-formatted string for a value, e.g. 0.9234 becomes 92.34% """

    return '{:,.2%}'.format(val)


def determine_boundedness(row):
    """ Determine the boundedness of a single row with topdown metrics """

    boundedness = []

    level_1 = max_column(row, ['retiring',
                               'bad_speculation',
                               'frontend_bound',
                               'backend_bound'])
    if str(row[level_1]) != 'nan' and str(row[level_1]) != 'inf':
        boundedness.append(level_1 + ' ' + percentage_string(row[level_1]))

    if level_1 == 'bad_speculation':
        level_2 = max_column(row, ['branch_mispredict',
                                   'machine_clear'])
        boundedness.append(level_2 + ' ' + percentage_string(row[level_2]))
    elif level_1 == 'frontend_bound':
        level_2 = max_column(row, ['frontend_latency',
                                   'frontend_bandwidth'])
        boundedness.append(level_2 + ' ' + percentage_string(row[level_2]))
    elif level_1 == 'backend_bound':
        level_2 = max_column(row, ['core_bound',
                                   'memory_bound'])
        boundedness.append(level_2 + ' ' + percentage_string(row[level_2]))
        if level_2 == 'memory_bound':
            level_3 = max_column(row, ['l1_bound',
                                       'l2_bound',
                                       'l3_bound',
                                       'mem_bound',
                                       'uncore_bound'])
            boundedness.append(level_3 + ' ' + percentage_string(row[level_3]))

    if len(boundedness) == 0:
        boundedness.append('undetermined')

    return boundedness


def analyze_topdown_metrics(dfm):
    """ Analyze topdown metrics to determine boundedness of different Caliper regions """

    dfm['boundedness'] = dfm.apply(determine_boundedness, axis=1)

    for metric in METRICS:
        del dfm[metric]

    return [dict([col for col in row.items() if str(col[1]) != 'nan'])
            for row in dfm.to_dict('index').values()]


def main():
    """ Print all Caliper entries with their derived metrics """

    if len(sys.argv) != 3:
        eprint("Usage: " + sys.argv[0] + " <json file> <arch, e.g. ivybridge>")

    dfm = pd.read_json(sys.argv[1])

    dfm = derive_topdown(dfm, sys.argv[2])

    analysis = analyze_topdown_metrics(dfm)

    for row in analysis:
        print(row)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2020, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
import os
import re
import sys
import math
from math import log, pi

import matplotlib.pyplot as plt
from matplotlib.pyplot import text
import numpy


GIGABYTE = 1.0e9
FONT_SIZE = 16
FONT_COLOR = "black"
FONT_WEIGHT = "bold"
WARP_SIZE = 32
DEBUG = (
    True
    if re.search(
        "^(t|y|on|[0-9])", os.environ.get("TIMEMORY_DEBUG", "off").lower()
    )
    is not None
    else False
)
VERBOSE = int(os.environ.get("TIMEMORY_VERBOSE", "0")) if not DEBUG else 255
RANK_INDEX = 0

__all__ = [
    "echo_dart_tag",
    "get_json_entry",
    "ert_params",
    "ert_counter",
    "ert_data",
    "smooth",
    "read_ert",
    "get_peak_ops",
    "get_peak_int_theo",
    "get_peak_bandwidth",
    "get_theo_bandwidth_txns",
    "get_hotspots",
    "get_hotspots_integer",
    "get_color",
    "plot_parameters",
    "plot_roofline",
    "FONT_SIZE",
    "VERBOSE",
    "DEBUG",
    "RANK_INDEX",
]


# -------------------------------------------------------------------------------------- #
#
def echo_dart_tag(name, filepath, img_type):
    """
    Printing this string will upload the results to CDash when running CTest
    """
    print(
        '<DartMeasurementFile name="{}" '.format(name)
        + 'type="image/{}">{}</DartMeasurementFile>'.format(img_type, filepath)
    )


# -------------------------------------------------------------------------------------- #
#
def get_json_entry(inp, key):
    for _key in [key, key.replace("_", "-"), key.replace("-", "_")]:
        if _key in inp:
            return inp[_key]
    return None


# -------------------------------------------------------------------------------------- #
#
def flatten_list(datalist):
    return [item for sublist in datalist for item in sublist]


# -------------------------------------------------------------------------------------- #
#
def get_font():
    return {
        "size": FONT_SIZE,
        "color": FONT_COLOR,
        "weight": FONT_WEIGHT,
        "family": "serif",
    }


# -------------------------------------------------------------------------------------- #
#
class ert_params:
    """
    ERT execution parameters
    """

    def __init__(self, inp):
        self.working_set_min = get_json_entry(inp, "working_set_min")
        self.memory_max = get_json_entry(inp, "memory_max")
        self.nthreads = get_json_entry(inp, "nthreads")
        self.nrank = get_json_entry(inp, "nrank")
        self.nproc = get_json_entry(inp, "nproc")
        self.nstreams = get_json_entry(inp, "nstreams")
        self.grid_size = get_json_entry(inp, "grid_size")
        self.block_size = get_json_entry(inp, "block_size")
        self.shmem_size = get_json_entry(inp, "shmem_size")

    def __str__(self):
        ret = []
        for itr in (
            "working_set_min",
            "memory_max",
            "nthreads",
            "nrank",
            "nproc",
            "nstreams",
            "grid_size",
            "block_size",
            "shmem_size",
        ):
            ret += ["=".join([itr, "{}".format(getattr(self, itr))])]
        return ", ".join(ret)


# -------------------------------------------------------------------------------------- #
#
class ert_counter:
    """
    ERT counter data
    """

    def __init__(self, inp):
        _data = get_json_entry(inp, "repr_data")
        self.units = get_json_entry(inp, "units")
        self.display_units = get_json_entry(inp, "display_units")
        if not isinstance(_data, list) and not isinstance(_data, tuple):
            self.data = [_data]
        else:
            self.data = _data
        if self.units is not None:
            for i in range(len(self.data)):
                self.data[i] /= self.units

    def __str__(self):
        ret = []
        for itr in ("data", "units", "display_units"):
            ret += ["=".join([itr, "{}".format(getattr(self, itr))])]
        return ", ".join(ret)

    def get(self, total):
        _data = self.data
        _list = []
        for i in range(len(_data)):
            _list.append(total / _data[i])
        return _list

    def get_warp_ops(
        self, total
    ):  # for int ops, scale the peak ops by 32 for warp-based measurements
        _data = (
            self.data
        )  # we can use the same function for transaction bandwidth
        _list = []
        for i in range(len(_data)):
            _list.append((total / _data[i]) / WARP_SIZE)
        return _list


# -------------------------------------------------------------------------------------- #
#
class ert_data:
    """
    ERT instance data
    """

    def __init__(self, inp):
        self.label = get_json_entry(inp, "label")
        self.working_set = get_json_entry(inp, "working-set")
        self.trials = get_json_entry(inp, "trials")
        self.total_bytes = get_json_entry(inp, "total-bytes")
        self.total_ops = get_json_entry(inp, "total-ops")
        self.ops_per_set = get_json_entry(inp, "ops-per-set")
        self.device = get_json_entry(inp, "device")
        self.dtype = get_json_entry(inp, "dtype")
        self.counter = ert_counter(get_json_entry(inp, "counter"))
        self.exec_params = ert_params(get_json_entry(inp, "exec-params"))

    def __str__(self):
        ret = []
        for itr in (
            "label",
            "working_set",
            "trials",
            "total_bytes",
            "total_ops",
            "ops_per_set",
            "device",
            "dtype",
        ):
            ret += ["=".join([itr, "{}".format(getattr(self, itr))])]
        for itr in ("counter", "exec_params"):
            ret += [
                "".join(
                    [
                        "\n    {:>12}".format(itr),
                        ": ({})".format(getattr(self, itr)),
                    ]
                )
            ]
        return ", ".join(ret)


# -------------------------------------------------------------------------------------- #
#
def smooth(x, y):
    """
    Smooth a curve
    """
    xs = x[:]
    ys = y[:]
    d = 0
    for i in range(0, len(ys)):
        num = min(len(ys), i + d + 1) - max(0, i - d)
        _beg = max(0, i - d)
        _end = min(len(ys), i + d + 1)
        _ys = ys[_beg:_end]
        total = sum(_ys)
        ys[i] = total / float(num)
    return xs, ys


# -------------------------------------------------------------------------------------- #
#
def read_ert(inp):
    data = inp

    def _read_ert(_data):
        if "ert" not in _data:
            return None
        else:
            inst = []
            _ert = _data["ert"]
            for element in _ert:
                inst.append(ert_data(element))
            return inst

    inst = _read_ert(data)
    if inst is None:
        if "roofline" in data:
            data = data["roofline"]
        inst = _read_ert(data)

    if VERBOSE > 2:
        for entry in inst:
            print("{}\n".format(entry))

    return inst


# -------------------------------------------------------------------------------------- #
#
def get_peak_ops(roof_data, flop_info=None):
    """
    Get the peak operations / sec
    """
    peak = {}

    for element in roof_data:
        total_ops = element.total_ops / GIGABYTE
        _label = element.label
        _data = element.counter.get(
            total_ops
        )  # this gives total_ops/repr_data (time) ops/sec
        if VERBOSE > 2:
            print("LABEL: {}, DATA: {}".format(_label, _data))
        if _label not in peak:
            peak[_label] = _data
        else:
            _peak = peak[_label]
            for i in range(len(_data)):
                if i >= len(_peak):
                    _peak += [_data[i]]
                else:
                    _peak[i] = max([_peak[i], _data[i]])

    info = "GFLOPs/sec"
    if flop_info is not None:
        try:
            info_list = re.sub(r"[^\w]", " ", flop_info).split()
            if len(info_list) > 0:
                info = info_list[0] + " GFLOPs/sec"
        except Exception:
            pass

    print("PEAK: {}".format(peak))

    peak_ops = [peak, info]
    return peak_ops


# -------------------------------------------------------------------------------------- #
#
def get_peak_int_theo(_inst_peak):
    peak = {}
    peak["Theoretical Peak"] = _inst_peak
    info = "warps GIPS"
    peak_ops = [peak, info]
    print(peak)
    return peak_ops


# -------------------------------------------------------------------------------------- #
#
def get_peak_bandwidth(roof_data, band_labels):
    """
    Get multi-level bandwidth peaks - Implementation from ERT:
    https://bitbucket.org/berkeleylab/cs-roofline-toolkit
    """
    ref_intensity = 0
    work_set = []
    bandwidth_data = []

    # Read bandwidth raw data
    for element in roof_data:
        intensity = element.ops_per_set
        if ref_intensity == 0:
            ref_intensity = intensity
        if intensity != ref_intensity:
            continue
        work_set.append(element.working_set)
        total_bytes = element.total_bytes / GIGABYTE
        bandwidth_data.append(element.counter.get(total_bytes))

    fraction = 1.05
    samples = 10000
    bandwidth_data = flatten_list(bandwidth_data)

    max_bandwidth = max(bandwidth_data)
    begin = bandwidth_data.index(max_bandwidth)

    work_set = work_set[begin:]
    bandwidth_data = bandwidth_data[begin:]
    # min_bandwidth = min(bandwidth_data)

    dband = max_bandwidth / float(samples - 1)

    counts = samples * [0]
    totals = samples * [0.0]

    work_set, bandwidth_data = smooth(work_set, bandwidth_data)

    for i in range(0, samples):
        cband = i * dband
        for bandwidth in bandwidth_data:
            if bandwidth >= cband / fraction and bandwidth <= cband * fraction:
                totals[i] += bandwidth
                counts[i] += 1

    band_list = [[1000 * max_bandwidth, 1000]]

    maxc = -1
    maxi = -1

    for i in range(samples - 3, 1, -1):
        if counts[i] > 10:
            if counts[i] > maxc:
                maxc = counts[i]
                maxi = i
        else:
            if maxc > 1:
                value = float(totals[maxi]) / max(1, counts[maxi])
                if 1.20 * value < float(band_list[-1][0]) / band_list[-1][1]:
                    band_list.append([totals[maxi], counts[maxi]])
                else:
                    band_list[-1][0] += totals[maxi]
                    band_list[-1][1] += counts[maxi]
            maxc = -1
            maxi = -1

    band_info_list = ["DRAM"]
    cache_num = len(band_list) - 1

    for cache in range(1, cache_num + 1):
        band_info_list = ["L%d" % (cache_num + 1 - cache)] + band_info_list

    peak_bandwidths = []
    band_info_list = [e for e in band_info_list if e in band_labels]

    for (band, band_info) in zip(band_list, band_info_list):
        band_info = band_info + " GB/s"
        peak_bandwidths.append([float(band[0] / band[1]), band_info])

    return peak_bandwidths


# -------------------------------------------------------------------------------------- #
#
def get_theo_bandwidth_txns(txn_bandwidth):

    peak_bandwidths = []
    peak_bandwidths.append([float(txn_bandwidth[0]), "L1 GTXNs/s"])
    peak_bandwidths.append([float(txn_bandwidth[1]), "L2 GTXNs/s"])
    peak_bandwidths.append([float(txn_bandwidth[2]), "DRAM GTXNs/s"])
    return peak_bandwidths


# -------------------------------------------------------------------------------------- #
#
def get_hotspots(op_data, ai_data, index=None):
    """
    Get the hotspots information
    """
    import itertools

    marker = itertools.cycle(("o", ",", "^", "+", "*", ">", "<"))

    if "type" not in op_data or "type" not in ai_data:
        return []

    op_data_type = op_data["type"]
    ai_data_type = ai_data["type"]

    op_type = None
    ai_type = None
    if "cpu_roofline" in op_data_type:
        op_type = "cpu"
    if "cpu_roofline" in ai_data_type:
        ai_type = "cpu"
    if "gpu_roofline" in op_data_type:
        op_type = "gpu"
    if "gpu_roofline" in ai_data_type:
        ai_type = "gpu"

    if index is None:
        index = RANK_INDEX

    if "ranks" in op_data:
        op_data = op_data["ranks"][index]
    if "ranks" in ai_data:
        ai_data = ai_data["ranks"][index]

    op_graph_data = op_data["graph"]
    ai_graph_data = ai_data["graph"]
    hotspots = []

    avg_runtime = 0.0
    max_runtime = 0.0
    max_length = min([len(op_graph_data), len(ai_graph_data)])
    all_runtime = []

    def get_runtime(_data, extra=[]):
        opts = ["runtime", "elapsed"] + extra
        for opt in opts:
            if opt in _data.keys():
                return float(_data[opt])
            elif opt.title() in _data.keys():
                return float(_data[opt.title()])
        # print("Could not find runtime field in {}. Searched: {}".format(_data, opts))
        return None

    def get_flops(_data, extra=[]):
        opts = [
            "flops",
            "counted_ops",
            "flop_count",
            "flop_count_sp",
            "flop_count_dp",
            "flop_count_hp",
            "DP_operations",
            "SP_operations",
        ] + extra
        for opt in opts:
            if opt in _data.keys():
                value = float(_data[opt])
                if value > 0.0:
                    return value
            elif opt.title() in _data.keys():
                value = float(_data[opt.title()])
                if value > 0.0:
                    return value
        # print("Could not find flops field in {}. Searched: {}".format(_data, opts))
        return None

    def get_bandwidth(_data, extra=[]):
        opts = [
            "bandwidth",
            "counted_ins",
            "ldst_executed",
            "L/S completed",
            "Loads_Stores_completed",
        ] + extra
        for opt in opts:
            if opt in _data.keys():
                return float(_data[opt])
            elif opt.title() in _data.keys():
                return float(_data[opt.title()])
        # print("Could not find bandwidth field in {}. Searched: {}".format(_data, opts))
        return None

    def check_ignore(_ai, _op):
        if (
            "cuptiOverhead" in _ai
            or "cudaRuntime" in _ai
            or "cuptiOverhead" in _op
            or "cudaRuntime" in _op
        ):
            return True
        return False

    for i in range(0, max_length):
        if check_ignore(ai_graph_data[i]["prefix"], op_graph_data[i]["prefix"]):
            continue

        ai_repr = ai_graph_data[i]["entry"]["repr_data"]
        op_repr = op_graph_data[i]["entry"]["repr_data"]
        all_runtime += filter(
            None, [get_runtime(ai_repr), get_runtime(op_repr)]
        )

    for rt in all_runtime:
        avg_runtime += rt

    if len(all_runtime) > 1:
        max_runtime = max(all_runtime)
        avg_runtime -= max_runtime
        avg_runtime /= len(all_runtime) - 1.0

    for i in range(0, max_length):
        if check_ignore(ai_graph_data[i]["prefix"], op_graph_data[i]["prefix"]):
            continue

        runtimes = []
        flop = None
        bandwidth = None

        ai_repr = ai_graph_data[i]["entry"]["repr_data"]
        op_repr = op_graph_data[i]["entry"]["repr_data"]

        label = op_graph_data[i]["prefix"]
        runtimes += filter(None, [get_runtime(ai_repr), get_runtime(op_repr)])

        flop = get_flops(op_repr)

        if flop is None:
            print("No flops found in {}!".format(op_repr))
            continue

        if op_type == "gpu" or ai_type == "gpu":
            bandwidth = get_bandwidth(op_repr)
        elif ai_type == "cpu" or op_type == "cpu":
            bandwidth = get_bandwidth(ai_repr)

        if bandwidth is None:
            bandwidth = get_bandwidth(op_repr)
            if bandwidth is None:
                bandwidth = get_bandwidth(ai_repr)

        runtime = 0.0
        for rt in runtimes:
            runtime += rt
        runtime /= len(runtimes)

        intensity = flop / bandwidth if bandwidth != 0.0 else 0.0
        flop = flop / GIGABYTE / runtime
        proportion = runtime / avg_runtime
        label = (
            re.sub(r"^[|0-9]+", " ", label)
            .replace("> [cxx] ", "")
            .replace("> [_c_] ", "")
            .replace("> [pyc] ", "")
        )

        if VERBOSE > 1:
            print(
                "intensity: {}, flop: {}, proportion: {}, label: {}".format(
                    intensity, flop, proportion, label
                )
            )

        if VERBOSE > 0:
            print(
                "{} : runtime = {}, avg = {}, proportion = {}, flop = {}".format(
                    label, runtime, avg_runtime, proportion, flop
                )
            )

        # this can arise from overflow
        if flop <= 1.0e-3 or bandwidth <= 0.0:
            continue

        my_marker = next(marker)
        hotspots.append([intensity, flop, proportion, my_marker, label])
    return hotspots


# -------------------------------------------------------------------------------------- #
#
def get_hotspots_integer(op_data, ai_data, index=None):
    """
    Get the hotspots information
    """
    import itertools

    marker = itertools.cycle(("o", ",", "^", "+", "*", ">", "<"))
    if "type" not in op_data or "type" not in ai_data:
        return []

    op_data_type = op_data["type"]

    op_type = None
    if "cpu_roofline" in op_data_type:
        op_type = "cpu"
    if "gpu_roofline" in op_data_type:
        op_type = "gpu"

    if index is None:
        index = RANK_INDEX

    if "ranks" in op_data:
        op_data = op_data["ranks"][index]
    if "ranks" in ai_data:
        ai_data = ai_data["ranks"][index]

    op_graph_data = op_data["graph"]
    ai_graph_data = ai_data["graph"]
    hotspots = []

    # avg_runtime = 1.0
    # max_runtime = 0.0
    # max_length = min([len(op_graph_data), len(ai_graph_data)])
    all_runtime = []

    def get_runtime(_data, extra=[]):
        opts = ["runtime", "elapsed"] + extra
        for opt in opts:
            if opt in _data:
                return float(_data[opt])
            elif opt.title() in _data:
                return float(_data[opt.title()])
        return None

    def get_int_ops_scaled(_data, extra=[]):
        opts = ["inst_integer"]
        for opt in opts:
            if opt in _data:
                value = float(_data[opt])
                if value > 0.0:
                    return (
                        value / WARP_SIZE
                    )  # scale it by 32 because its a thread level counter
        return None

    def get_HBM_transactions(
        _data, extra=[]
    ):  # each transaction is 32 bytes long
        _HBM_transactions = 0
        opts = ["dram_read_transactions", "dram_write_transactions"] + extra
        for opt in opts:
            _HBM_transactions += float(_data[opt])
        if _HBM_transactions > 0.0:
            return _HBM_transactions

        else:
            print("No Memory Transactions counted")
            return 0

    def get_L2_transactions(_data, extra=[]):
        _L2_transactions = 0
        opts = ["l2_read_transactions", "l2_write_transactions"] + extra
        for opt in opts:
            _L2_transactions += float(_data[opt])
        if _L2_transactions > 0.0:
            return _L2_transactions
        else:
            return 0

    def get_L1_transactions(_data, extra=[]):
        _L1_transactions = 0
        opts = [
            "gld_transactions",
            "gst_transactions",
            "shared_store_transactions",
            "shared_load_transactions",
        ] + extra
        for opt in opts:
            if "shared" in opt:
                _L1_transactions += float(_data[opt]) * 4
            else:
                _L1_transactions += float(_data[opt])
        if _L1_transactions > 0.0:
            return _L1_transactions
        else:
            return 0

    def check_ignore(_ai, _op):
        if (
            "cuptiOverhead" in _ai
            or "cudaRuntime" in _ai
            or "cuptiOverhead" in _op
            or "cudaRuntime" in _op
        ):
            return True
        return False

    all_runtime = {}
    for i in range(0, len(ai_graph_data)):
        ai_repr = ai_graph_data[i]["entry"]["repr_data"]
        all_runtime[ai_graph_data[i]["hash"]] = get_runtime(ai_repr)

    for i in range(0, len(op_graph_data)):

        Iops = None
        HBM_transactions = None
        L1_transactions = None
        L2_transactions = None

        op_repr = op_graph_data[i]["entry"]["repr_data"]

        label = op_graph_data[i]["prefix"]
        if "thrust" in label:
            label = "thrust kernel"
        else:
            label = label.split("(")[0]

        if op_type == "gpu":
            Iops = get_int_ops_scaled(op_repr)
            HBM_transactions = get_HBM_transactions(op_repr)
            if HBM_transactions == 0:
                continue

            L1_transactions = get_L1_transactions(op_repr)
            if L1_transactions == 0:
                continue
            L2_transactions = get_L2_transactions(op_repr)
            if L2_transactions == 0:
                continue
        else:
            print("no GPU data available")

        runtime = 0.0
        runtime = all_runtime[op_graph_data[i]["hash"]]

        # hotspot for HBM
        HBM_intensity = Iops / HBM_transactions  # if bandwidth != 0.0 else 0.0
        HBM_GIPS = Iops / GIGABYTE / runtime
        HBM_proportion = 0.1  # runtime / avg_runtime

        # hotspot for L1
        L1_intensity = Iops / L1_transactions  # if bandwidth != 0.0 else 0.0
        L1_GIPS = Iops / GIGABYTE / runtime
        L1_proportion = 0.5  # runtime / avg_runtime

        # hotspot for L2
        L2_intensity = Iops / L2_transactions  # if bandwidth != 0.0 else 0.0
        L2_GIPS = Iops / GIGABYTE / runtime
        L2_proportion = 1  # runtime / avg_runtime

        # label = re.sub(r'^[|0-9]+', ' ', label).replace("> [cxx] ", "").replace(
        #    "> [_c_] ", "").replace("> [pyc] ", "")

        # if VERBOSE > 1:
        #     print("intensity: {}, flop: {}, proportion: {}, label: {}".format(
        #         intensity, flop, proportion, label))
        #
        # if VERBOSE > 0:
        #     print("{} : runtime = {}, avg = {}, proportion = {}, flop = {}".format(
        #         label, runtime, avg_runtime, proportion, flop))
        #
        # # this can arise from overflow
        # if flop <= 1.0e-3 or bandwidth <= 0.0:
        #     continue

        my_marker = next(marker)
        hotspots.append(
            [
                HBM_intensity,
                HBM_GIPS,
                HBM_proportion,
                my_marker,
                "HBM: " + label,
            ]
        )
        hotspots.append(
            [L1_intensity, L1_GIPS, L1_proportion, my_marker, "L1: " + label]
        )
        hotspots.append(
            [L2_intensity, L2_GIPS, L2_proportion, my_marker, "L2: " + label]
        )
    return hotspots


# -------------------------------------------------------------------------------------- #
#
def get_color(proportion):
    if proportion < 0.2:
        color = "green"
    elif proportion < 0.6:
        color = "blue"
    else:
        color = "red"
    return color


# -------------------------------------------------------------------------------------- #
#
class plot_parameters:
    def __init__(self, peak_flops, hotspots, _inst_roofline):

        _peak = peak_flops[0]
        if isinstance(_peak, list):
            _peak = max(_peak)
        elif isinstance(_peak, dict):
            tmp = 0.0
            for key, entry in _peak.items():
                tmp = max([tmp] + entry)
            _peak = tmp

        y_digits = int(math.log10(_peak)) + 1
        self.xmin = 0.01
        self.xmax = 1000
        self.ymin = 1
        self.ymax = 10 ** y_digits

        if _inst_roofline:
            self.xlabel = "Instruction Intensity [Warp Inst. Per Transaction]"
            self.ylabel = "Performance [Warp GIPS]"
        else:
            self.xlabel = "Arithmetic Intensity [FLOPs/Byte]"
            self.ylabel = "Performance [GFLOPs/sec]"

        for element in hotspots:
            intensity = element[0]
            flop = element[1]
            if flop > self.ymax:
                self.ymax = 10 ** int(log(flop) / log(10) + 1)
            if flop < self.ymin:
                self.ymin = 10 ** int(log(flop) / log(10) - 1)
            if intensity > self.xmax:
                self.xmax = 100 ** int(log(intensity) / log(10) + 1)
            if intensity < self.xmin:
                self.xmin = 10 ** int(log(intensity) / log(10) - 1)

        print(
            "X (min, max) = {}, {}, Y (min, max) = {}, {}".format(
                self.xmin, self.xmax, self.ymin, self.ymax
            )
        )


#
# -------------------------------------------------------------------------------------- #
#
def plot_roofline(
    ai_data,
    op_data,
    *_args,
    **_kwargs,
):
    _ai = []
    _op = []

    if "timemory" in ai_data:
        for key, data in ai_data["timemory"].items():
            _ai.append(data)
    else:
        _ai.append(ai_data)

    if "timemory" in op_data:
        for key, data in op_data["timemory"].items():
            _op.append(data)
    else:
        _op.append(op_data)

    for ai, op in zip(_ai, _op):
        plot_roofline_impl(ai, op, *_args, **_kwargs)


# -------------------------------------------------------------------------------------- #
#
def plot_roofline_impl(
    ai_data,
    op_data,
    band_labels,
    txn_bandwidths,
    inst_peak,
    _rtype,
    display=False,
    fname="roofline",
    image_type="png",
    output_dir=os.getcwd(),
    title="Roofline Plot",
    width=1600,
    height=1200,
    dpi=100,
    echo_dart=False,
    index=None,
):
    """Plot the roofline"""

    if index is None:
        index = RANK_INDEX

    inst_roofline = False
    if "gpu_roofline_inst" in _rtype:
        inst_roofline = True
        print("GPU INST ROOFLINE ON")
    else:
        print("GPU INST ROOFLINE OFF")

    band_data = read_ert(ai_data)
    peak_data = read_ert(op_data)

    info = op_data["unit_repr"] if "unit_repr" in op_data else None

    if inst_roofline:
        peak_flop = get_peak_int_theo(inst_peak)
    else:
        peak_flop = get_peak_ops(peak_data, info)

    if inst_roofline:
        peak_band = get_theo_bandwidth_txns(txn_bandwidths)
    else:
        peak_band = get_peak_bandwidth(band_data, band_labels)

    if inst_roofline:
        hotspots = get_hotspots_integer(op_data, ai_data, index)
    else:
        hotspots = get_hotspots(op_data, ai_data, index)

    print("peak_flop = {}, peak_band = {}".format(peak_flop, peak_band))

    plot_params = plot_parameters(peak_flop, hotspots, inst_roofline)

    f = plt.figure(figsize=(width / dpi, height / dpi), dpi=dpi)
    f.add_subplot(111)

    _title_font = get_font()
    _title_font["size"] += 8
    plt.title(title.title(), **_title_font)
    plt.grid(True, which="major", ls="--", lw=1)
    plt.grid(True, which="minor", ls="--", lw=0.5)
    plt.yscale("log")
    plt.xscale("log")

    plt.xlabel(plot_params.xlabel, **get_font())
    plt.ylabel(plot_params.ylabel, **get_font())

    axes = plt.gca()
    axes.set_xlim([plot_params.xmin, plot_params.xmax])
    axes.set_ylim([plot_params.ymin, plot_params.ymax])

    # _peak_flop = peak_flop[0]
    # if isinstance(_peak_flop, list):
    #    _peak_flop = max(_peak_flop)

    # plot bandwidth roof
    _nitr = 0
    for _label, _peak_flop in peak_flop[0].items():
        x0 = plot_params.xmax
        print("Label: {}, Peak: {}".format(_label, _peak_flop))
        _peakop = max(_peak_flop)
        for band in peak_band:
            x1 = plot_params.xmin
            y1 = band[0] * plot_params.xmin
            if y1 < plot_params.ymin:
                x1 = plot_params.ymin / band[0]
                y1 = plot_params.ymin
            x2 = _peakop / band[0]
            y2 = _peakop
            if x2 < x0:
                x0 = x2

            x1log = log(x1) / log(10)
            x2log = log(x2) / log(10)
            y1log = log(y1) / log(10)
            y2log = log(y2) / log(10)
            x_text = 10 ** ((x1log + x2log) / 2)
            y_text = 10 ** ((y1log + y2log) / 2)

            fig = plt.gcf()
            size = fig.get_size_inches() * fig.dpi
            fig_x, fig_y = size

            dx = log(x2) - log(x1)
            dy = log(y2) - log(y1)
            x_min, x_max = plt.xlim()
            y_min, y_max = plt.ylim()
            Dx = dx * fig_x / (log(x_max) - log(x_min))
            Dy = dy * fig_y / (log(y_max) - log(y_min))
            fdiv = 1.0
            angle = (180.0 / pi) * numpy.arctan(Dy / Dx / fdiv)

            if _nitr == 0:
                text(
                    x_text,
                    y_text,
                    "%.2f %s" % (band[0], band[1]),
                    rotation=angle,
                    rotation_mode="anchor",
                    **get_font(),
                )
            plt.plot([x1, x2], [y1, y2], color="magenta")

            # plot computing roof
            temp_label = _label.replace("_", "-")
            temp_label = temp_label.upper()
            text(
                plot_params.xmax,
                _peakop + 2,
                "%.2f %s %s" % (_peakop, temp_label, peak_flop[1]),
                horizontalalignment="right",
                **get_font(),
            )

            plt.plot([x0, plot_params.xmax], [_peakop, _peakop], color="b")
        # ensure bandwidth labels are not duplicated
        _nitr += 1

    # from random import random
    import pylab

    labels = []
    # plot hotspots
    plotted_spots = []
    for element in hotspots:
        if VERBOSE > 0:
            print(element[0], element[1])
        c = get_color(element[2])
        myScatter = plt.scatter(element[0], element[1], c=c, marker=element[3])
        plotted_spots.append(myScatter)
        # factor = 10.0 * random() - 5.0
        label = "{}".format(element[4]).replace(">>>", "")

        # plt.annotate(label, (element[0], element[1] + factor), **get_font())
        labels.append(label)
        # text(element[0], element[1], "%s" %
        #     element[3], rotation=0, rotation_mode='anchor')

    # try:
    #    from adjustText import adjust_text
    print(labels)
    try:
        pylab.legend(
            plotted_spots,
            labels,
            prop={"size": (FONT_SIZE - 2)},
            bbox_to_anchor=(1.04, 1),
            loc="upper left",
        )
    except Exception as e:
        sys.stderr.write(f"{e}\n")
        pylab.legend(
            plotted_spots,
            labels,
            prop={"size": (FONT_SIZE - 2)},
        )

    # plt.legend(labels)
    if display:
        print("Displaying plot...")
        plt.show()
    else:
        imgfname = os.path.join(output_dir, os.path.basename(fname))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        if not ".{}".format(image_type) in imgfname:
            imgfname += ".{}".format(image_type)
        print('Saving plot: "{}"...'.format(imgfname))
        plt.savefig(imgfname, bbox_inches="tight")
        plt.close()
        if echo_dart:
            bfname = os.path.basename(fname)
            echo_dart_tag(bfname, imgfname, image_type)



# Thread tests

import unittest

import calipertest as calitest

class CaliperThreadTest(unittest.TestCase):
    """ Caliper thread test case """

    def test_thread(self):
        target_cmd = [ './ci_test_thread' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'thread-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) >= 20)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, {'local', 'global', 'function'}))
        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, {'pthread.id', 'pthread.is_master', 'my_thread_id', 'global', 'function'}))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, {'my_thread_id' : '49',
                        'pthread.is_master' : 'false',
                        'function'     : 'thread_proc',
                        'global'       : '999' }))
        self.assertTrue(calitest.has_snapshot_with_attributes(
            snapshots, { 'function'    : 'main',
                         'pthread.is_master' : 'true',
                         'local'       : '99' }))

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, {'pthread.id', 'my_thread_id', 'global', 'pthread_exp'}))
        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, {'global', 'my_thread_id', 'nopthread_exp'}))
        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, {'event.set#pthread.id', 'pthread_exp'}))
        self.assertFalse(calitest.has_snapshot_with_keys(
            snapshots, {'event.set#pthread.id', 'nopthread_exp'}))

if __name__ == "__main__":
    unittest.main()



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Muhammad Haseeb"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Muhammad Haseeb"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import time
import mmap
import numpy as np
import unittest
import random
import tempfile
import timemory as tim
from timemory import component as comp

# --------------------------- test setup variables ----------------------------------- #
niter = 20
nelements = 0.95 * (mmap.PAGESIZE * 500)
memory_unit = (tim.units.kilobyte, "KiB")

peak = comp.PeakRss("peak_rss")
curr = comp.PageRss("curr_rss")
rb = comp.ReadBytes("read_bytes")
wb = comp.WrittenBytes("written_bytes")
current_peak = comp.CurrentPeakRss("curr_peak_rss")

sizeof_int64_t = np.ones(1, np.int64).itemsize
sizeof_double = np.ones(1, np.float64).itemsize

tot_size = nelements * sizeof_int64_t / memory_unit[0]
tot_rw = nelements * sizeof_double / memory_unit[0]

peak_tolerance = 5 * tim.units.megabyte
curr_tolerance = 5 * tim.units.megabyte
byte_tolerance = tot_rw


# --------------------------- helper functions ----------------------------------------- #
# compute fibonacci
def fibonacci(n):
    return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


# get a random entry
def random_entry(vect):
    random.seed(int(1.0e9 * time.time()))
    num = int(random.uniform(0, len(vect) - 1))
    return vect[num]


# allocate memory
def allocate():
    peak.reset()
    curr.reset()
    current_peak.reset()

    curr.start()
    peak.start()
    current_peak.start()

    v = np.full(int(nelements), 15, dtype=np.int64)
    ret = fibonacci(0)
    nfib = random_entry(v)

    for i in range(niter):
        nfib = random_entry(v)
        ret += fibonacci(nfib)

    if ret < 0:
        print("fibonacci({}) * {} = {}\n".format(nfib, niter, ret))

    current_peak.stop()
    curr.stop()
    peak.stop()


# read and write to file
def read_write():

    rb.start()
    wb.start()

    ofs = tempfile.NamedTemporaryFile(delete=False)
    fname = ofs.name

    writeArray = np.array(range(int(nelements)))
    writeArray = writeArray.astype(np.float64)

    # make a new bytes array
    writeByteArray = bytearray(writeArray)

    # write the data to file
    ofs.write(writeByteArray)

    # close the file
    ofs.close()

    # read the data from file
    data = np.fromfile(fname, dtype=np.float64)

    rb.stop()
    wb.stop()

    return data


# print_info for peak, page rss
def print_info(name, obj, expected):
    print("")
    print("[" + name + "]>  measured :", obj.get())
    print("[" + name + "]>  expected :", expected, memory_unit[1])


# print_info for wb and rb
def print_info_wb_rb(name, obj, expected):
    print("")
    print("[" + name + "]>  measured :", obj.get()[0])
    print("[" + name + "]>  expected :", expected, memory_unit[1])


# print_info for current peak
def print_info_curr_peak(name, obj, expected):
    print("")
    print("[" + name + "]>  measured :", obj.get()[1] - obj.get()[0])
    print("[" + name + "]>  expected :", expected, memory_unit[1])


# --------------------------- RUsage Tests set ---------------------------------------- #
# RSS usage tests class
class TimemoryRUsageTests(unittest.TestCase):
    # setup class: timemory settings
    @classmethod
    def setUpClass(self):
        tim.settings.verbose = 0
        tim.settings.debug = False
        tim.settings.json_output = True
        tim.settings.precision = 9
        tim.settings.memory_units = memory_unit[1]
        tim.settings.mpi_thread = False
        tim.settings.file_output = False

        # perform allocation once here
        allocate()

        # perform read write once here
        read_write()

        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False
        tim.settings.dart_type = "peak_rss"

    # tear down class: finalize
    @classmethod
    def tearDownClass(self):
        # timemory finalize
        # tim.finalize()
        # tim.dmp.finalize()
        pass

    # ---------------------------------------------------------------------------------- #
    # test peak rss
    def test_peak_rss(self):
        """peak rss"""
        if not comp.PeakRss.available:
            raise unittest.SkipTest("[{}] not available".format("peak_rss"))

        print_info(self.shortDescription(), peak, tot_size)
        self.assertAlmostEqual(tot_size, peak.get(), delta=peak_tolerance)

    # ---------------------------------------------------------------------------------- #
    # test page rss
    def test_page_rss(self):
        """page rss"""
        if not comp.PageRss.available:
            raise unittest.SkipTest("[{}] not available".format("page_rss"))

        print_info(self.shortDescription(), curr, tot_size)
        self.assertAlmostEqual(tot_size, curr.get(), delta=curr_tolerance)

    # ---------------------------------------------------------------------------------- #
    # test read_bytes
    def test_read_bytes(self):
        """read bytes"""
        if not comp.ReadBytes.available:
            raise unittest.SkipTest("[{}] not available".format("read_bytes"))

        print_info_wb_rb(self.shortDescription(), rb, tot_rw)
        self.assertAlmostEqual(tot_rw, rb.get()[0], delta=byte_tolerance)

    # ---------------------------------------------------------------------------------- #
    # test write_bytes
    def test_written_bytes(self):
        """written bytes"""
        if not comp.WrittenBytes.available:
            raise unittest.SkipTest(
                "[{}] not available".format("written_bytes")
            )

        print_info_wb_rb(self.shortDescription(), wb, tot_rw)
        self.assertAlmostEqual(tot_rw, wb.get()[0], delta=byte_tolerance)

    # ---------------------------------------------------------------------------------- #
    # test write_bytes
    def test_current_peak_rss(self):
        """current peak rss"""
        if not comp.CurrentPeakRss.available:
            raise unittest.SkipTest(
                "[{}] not available".format("current_peak_rss")
            )

        print_info_curr_peak(self.shortDescription(), current_peak, tot_size)
        self.assertAlmostEqual(
            tot_size,
            current_peak.get()[1] - current_peak.get()[0],
            delta=peak_tolerance,
        )


# ----------------------------- main test runner -------------------------------------- #
# test runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#!/usr/bin/env python3

def _makelist(elem_or_list):
    if isinstance(elem_or_list, list):
        return elem_or_list
    else:
        return [ elem_or_list ]

def print_record(rec, ctx_key='path', metric_key='avg#inclusive#sum#time.duration'):
    if ctx_key in rec and metric_key in rec:
        print("  {0}: {1}s".format('/'.join(_makelist(rec[ctx_key])), rec[metric_key]))

#
# Use read_caliper_contents() to read the Caliper file
#

print("Using read_caliper_contents():")

from caliperreader.caliperreader import read_caliper_contents

records, globals = read_caliper_contents('example-profile.cali')

print("Globals:")
print("  globals['cali.caliper.version']: {0}".format(globals['cali.caliper.version']))
print("  globals['problem_size' {0}".format(globals['problem_size']))

print('\nMetrics:')
for rec in records:
    print_record(rec)


#
# Use a CaliperReader object to read the Caliper file
#

print ("\n\nUsing CaliperReader object:")

from caliperreader.caliperreader import CaliperReader

r = CaliperReader()

r.read('example-profile.cali')

print("Globals:")
print("  r.globals['cali.channel']: {0}".format(r.globals['cali.channel']))
print("  r.globals['figure_of_merit']: {0}".format(r.globals['figure_of_merit']))

print('\nMetrics:')
for rec in r.records:
    print_record(rec)

# We can query additional metadata about the Caliper attributes:
print("\nAttribute metadata:")
print("  r.attribute('function').is_nested(): {0}".format(r.attribute('function').is_nested()))
print("  r.attribute('function').attribute_type(): {0}".format(r.attribute('function').attribute_type()))
print("  r.attribute('figure_of_merit').get('adiak.type'): {0}".format(r.attribute('figure_of_merit').get('adiak.type')))

#
# Use CaliperStreamReader
#

print ("\n\nUsing CaliperStreamReader object:")

from caliperreader.caliperstreamreader import CaliperStreamReader

sr = CaliperStreamReader()

print('\nMetrics:')
sr.read('example-profile.cali', print_record)



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums


def generate_enum(component, ncount, indent):
    """
    This function generates a type trait label for C++
    """
    enumeration = mangled_enums.get(component, component)
    return "{}{:30} = {},".format(" " * indent, enumeration.upper(), ncount)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--verbose", help="Enable verbosity", default=0, type=int)
    parser.add_argument("-i", "--indent", help="Indentation", default=4, type=int)
    args = parser.parse_args()

    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))

    components.sort()
    outdata = "enum TIMEMORY_NATIVE_COMPONENT\n{\n"
    for i in range(0, len(components)):
        outdata += "{}\n".format(generate_enum(components[i], i, args.indent))

    outdata += "{}{} {} = \n{}    ({} + {})\n".format(" " * args.indent,
                                       "TIMEMORY_USER_COMPONENT_ENUM",
                                       "TIMEMORY_COMPONENTS_END", " " * args.indent,
                                       "TIMEMORY_COMPONENT_ENUM_SIZE",
                                       "TIMEMORY_USER_COMPONENT_ENUM_SIZE")
    outdata += "};"
    print("{}".format(outdata))



#!@PYTHON_EXECUTABLE@
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# @file util.py
# Decorators for timemory module
#

from __future__ import absolute_import
import six
import inspect
from enum import Enum
from functools import wraps

from ..common import FILE, FUNC, LINE, FRAME
from ..libs.libpytimemory import timer_decorator as _timer_decorator
from ..libs.libpytimemory import component_decorator as _component_decorator
from ..libs.libpytimemory import timer as _timer
from ..libs.libpytimemory import rss_usage as _rss_usage
from ..libs.libpytimemory import component as _component

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"
__all__ = [
    "base_decorator",
    "auto_timer",
    "auto_tuple",
    "timer",
    "rss_usage",
    "marker",
]


class context(Enum):
    blank = 0
    basic = 1
    full = 2
    defer = 3


class base_decorator(object):
    """
    A base class for the decorators and context managers
    """

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(self, key="", add_args=False, is_class=False, mode="defer"):
        self.key = key
        self.add_args = add_args
        self.is_class = is_class
        self.signature = context.defer
        self.mode = mode
        if self.mode != "defer":
            try:
                self.signature = getattr(context, self.mode)
            except AttributeError:
                pass

    # ---------------------------------------------------------------------------------- #
    #
    def determine_signature(self, is_decorator=True, is_context_manager=False):
        def determine_deferred(_is_decorate, _is_context):
            if _is_decorate:
                return context.basic
            elif _is_context:
                return context.blank
            else:
                return context.full

        if self.mode != "defer":
            try:
                self.signature = getattr(context, self.mode)
            except Exception as e:
                print(e)
                self.signature = determine_deferred(
                    is_decorator, is_context_manager
                )
        elif self.signature == context.defer:
            self.signature = determine_deferred(
                is_decorator, is_context_manager
            )

    # ---------------------------------------------------------------------------------- #
    #
    def parse_wrapped(self, func, args, kwargs):

        if (
            len(args) > 0
            and args[0] is not None
            and inspect.isclass(type(args[0]))
        ):
            self.is_class = True
        else:
            self.is_class = False

    # ---------------------------------------------------------------------------------- #
    #
    def class_string(self, args, kwargs):
        """
        Generate a class identifier
        """
        _str = ""
        if self.is_class and len(args) > 0 and args[0] is not None:
            _str = "[{}]".format(type(args[0]).__name__)
            # this check guards against old bug from class methods
            # calling class methods
            if _str not in self.key:
                return _str
            else:
                return ""
        return _str

    # ---------------------------------------------------------------------------------- #
    #

    def arg_string(self, _frame):
        """
        Generate a string of the arguments
        """
        # _str = '{}'.format(self.class_string(args, kwargs))
        _str = ""
        if self.add_args:
            _str = "{}".format(
                inspect.formatargvalues(*inspect.getargvalues(_frame))
            )
        return _str


#
class auto_timer(base_decorator):
    """A decorator or context-manager for the auto-timer, e.g.:
    @timemory.util.auto_timer(add_args=True)
    def main(n=5):
        for i in range(2):
            fibonacci(n * (i+1))
    # ...
    # output :
    # > [pyc] main(5)@'example.py':10 ...
    """

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(
        self,
        key="",
        add_args=False,
        is_class=False,
        report_at_exit=False,
        mode="defer",
    ):
        super(auto_timer, self).__init__(
            key=key, add_args=add_args, is_class=is_class, mode=mode
        )
        self.report_at_exit = report_at_exit
        self._self_obj = None

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """
        Decorator
        """
        _file = FILE(3)
        _line = LINE(2)

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            self.parse_wrapped(func, args, kwargs)
            self.determine_signature(
                is_decorator=True, is_context_manager=False
            )

            _frame = FRAME(1)
            _func = func.__name__
            _key = ""
            _args = self.arg_string(_frame)
            if self.signature == context.blank:
                _key = "{}{}".format(self.key, _args)
            elif self.signature == context.basic:
                _key = "{}{}/{}".format(_func, _args, self.key)
            elif self.signature == context.full:
                _key = "{}{}@{}:{}/{}".format(
                    _func, _args, _file, _line, self.key
                )
            _key = _key.strip("/")

            _dec = _timer_decorator(_key, self.report_at_exit)
            _ret = func(*args, **kwargs)
            del _dec
            return _ret

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """
        Context manager
        """
        _file = FILE(3)
        _line = LINE(2)
        _func = FUNC(2)
        _frame = FRAME(2)
        self.determine_signature(is_decorator=False, is_context_manager=True)

        _key = ""
        _args = self.arg_string(_frame)
        if self.signature == context.blank:
            _key = "{}{}".format(self.key, _args)
        elif self.signature == context.basic:
            _key = "{}{}/{}".format(_func, _args, self.key)
        elif self.signature == context.full:
            _key = "{}{}@{}:{}/{}".format(_func, _args, _file, _line, self.key)
        _key = _key.strip("/")

        self._self_obj = _timer_decorator(_key, self.report_at_exit)

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exc_type, exc_value, exc_traceback):
        del self._self_obj

        if (
            exc_type is not None
            and exc_value is not None
            and exc_traceback is not None
        ):
            import traceback

            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=5
            )


#
class timer(base_decorator):
    """A decorator or context-manager for the timer, e.g.:

    class someclass(object):

        @timemory.util.timer()
        def __init__(self):
            self.some_obj = None
    # ...
    """

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(self, key="", add_args=False, is_class=False, mode="defer"):
        super(timer, self).__init__(
            key=key, add_args=add_args, is_class=is_class, mode=mode
        )
        self._self_obj = None

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """
        Decorator
        """
        _file = FILE(3)
        _line = LINE(2)

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            self.parse_wrapped(func, args, kwargs)
            self.determine_signature(
                is_decorator=True, is_context_manager=False
            )

            _frame = FRAME(1)
            _func = func.__name__
            _key = ""
            _args = self.arg_string(_frame)
            if self.signature == context.blank:
                _key = "{}{}".format(self.key, _args)
            elif self.signature == context.basic:
                _key = "{}{}/{}".format(_func, _args, self.key)
            elif self.signature == context.full:
                _key = "{}{}@{}:{}/{}".format(
                    _func, _args, _file, _line, self.key
                )
            _key = _key.strip("/")

            t = _timer(_key)

            t.start()
            ret = func(*args, **kwargs)
            t.stop()
            t.report()
            return ret

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """
        Context manager
        """
        _file = FILE(3)
        _line = LINE(2)
        _func = FUNC(2)
        _frame = FRAME(2)
        self.determine_signature(is_decorator=False, is_context_manager=True)

        _key = ""
        _args = self.arg_string(_frame)
        if self.signature == context.blank:
            _key = "{}{}".format(self.key, _args)
        elif self.signature == context.basic:
            _key = "{}{}/{}".format(_func, _args, self.key)
        elif self.signature == context.full:
            _key = "{}{}@{}:{}/{}".format(_func, _args, _file, _line, self.key)
        _key = _key.strip("/")

        self._self_obj = _timer(_key)
        self._self_obj.start()

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._self_obj.stop()
        self._self_obj.report()

        if (
            exc_type is not None
            and exc_value is not None
            and exc_traceback is not None
        ):
            import traceback

            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=5
            )


#
class rss_usage(base_decorator):
    """A decorator or context-manager for the rss usage, e.g.:

    class someclass(object):

        @timemory.util.rss_usage()
        def __init__(self):
            self.some_obj = None
    # ...
    """

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(self, key="", add_args=False, is_class=False, mode="defer"):
        super(rss_usage, self).__init__(
            key=key, add_args=add_args, is_class=is_class, mode=mode
        )
        self._self_obj = None
        self._self_dif = None

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """
        Decorator
        """
        _file = FILE(3)
        _line = LINE(2)

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            self.parse_wrapped(func, args, kwargs)
            self.determine_signature(
                is_decorator=True, is_context_manager=False
            )

            _frame = FRAME(1)
            _func = func.__name__
            _key = ""
            _args = self.arg_string(_frame)
            if self.signature == context.blank:
                _key = "{}{}".format(self.key, _args)
            elif self.signature == context.basic:
                _key = "{}{}/{}".format(_func, _args, self.key)
            elif self.signature == context.full:
                _key = "{}{}@{}:{}/{}".format(
                    _func, _args, _file, _line, self.key
                )
            _key = _key.strip("/")

            self._self_obj = _rss_usage(_key)
            self._self_dif = _rss_usage(_key)
            self._self_dif.record()
            # run function
            ret = func(*args, **kwargs)
            # record
            self._self_obj.record()
            self._self_obj -= self._self_dif
            print("{}".format(self._self_obj))

            return ret

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """
        Context manager entrance
        """
        _file = FILE(3)
        _line = LINE(2)
        _func = FUNC(2)
        _frame = FRAME(2)
        self.determine_signature(is_decorator=False, is_context_manager=True)

        _key = ""
        _args = self.arg_string(_frame)
        if self.signature == context.blank:
            _key = "{}{}".format(self.key, _args)
        elif self.signature == context.basic:
            _key = "{}{}/{}".format(_func, _args, self.key)
        elif self.signature == context.full:
            _key = "{}{}@{}:{}/{}".format(_func, _args, _file, _line, self.key)
        _key = _key.strip("/")

        self._self_obj = _rss_usage(_key)
        self._self_dif = _rss_usage(_key)
        self._self_dif.record()

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Context manager exit
        """
        self._self_obj.record()
        self._self_obj -= self._self_dif
        print("{}".format(self._self_obj))

        if (
            exc_type is not None
            and exc_value is not None
            and exc_traceback is not None
        ):
            import traceback

            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=5
            )


#
class marker(base_decorator):
    """A decorator or context-manager for a marker, e.g.:
    @timemory.util.marker(components=("wall_clock", timemory.component.peak_rss))
    def main(n=5):
        for i in range(2):
            fibonacci(n * (i+1))
    # ...
    # output :
    # > [pyc] main(5)@'example.py':10 ...
    """

    # ---------------------------------------------------------------------------------- #
    #
    @staticmethod
    def get_components(items):
        ret = []
        for x in items:
            if isinstance(x, six.string_types):
                try:
                    ret.append(getattr(_component, x))
                except AttributeError:
                    pass
            else:
                ret.append(x)
        return ret

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(
        self,
        components=[],
        key="",
        add_args=False,
        is_class=False,
        report_at_exit=False,
        mode="defer",
    ):
        super(marker, self).__init__(
            key=key, add_args=add_args, is_class=is_class, mode=mode
        )
        self.components = marker.get_components(components)
        self.report_at_exit = report_at_exit
        self._self_obj = None

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """
        Decorator
        """
        _file = FILE(3)
        _line = LINE(2)

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            self.parse_wrapped(func, args, kwargs)
            self.determine_signature(
                is_decorator=True, is_context_manager=False
            )

            _frame = FRAME(1)
            _func = func.__name__
            _key = ""
            _args = self.arg_string(_frame)
            if self.signature == context.blank:
                _key = "{}{}".format(self.key, _args)
            elif self.signature == context.basic:
                _key = "{}{}/{}".format(_func, _args, self.key)
            elif self.signature == context.full:
                _key = "{}{}@{}:{}/{}".format(
                    _func, _args, _file, _line, self.key
                )
            _key = _key.strip("/")

            t = _component_decorator(self.components, _key)
            ret = func(*args, **kwargs)
            del t
            return ret

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """
        Context manager
        """
        _file = FILE(3)
        _line = LINE(2)
        _func = FUNC(2)
        _frame = FRAME(2)
        self.determine_signature(is_decorator=False, is_context_manager=True)

        _key = ""
        _args = self.arg_string(_frame)
        if self.signature == context.blank:
            _key = "{}{}".format(self.key, _args)
        elif self.signature == context.basic:
            _key = "{}{}/{}".format(_func, _args, self.key)
        elif self.signature == context.full:
            _key = "{}{}@{}:{}/{}".format(_func, _args, _file, _line, self.key)
        _key = _key.strip("/")

        self._self_obj = _component_decorator(self.components, _key)

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exc_type, exc_value, exc_traceback):
        del self._self_obj

        if (
            exc_type is not None
            and exc_value is not None
            and exc_traceback is not None
        ):
            import traceback

            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=5
            )


auto_tuple = marker



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import
from __future__ import division

# flake8: noqa

import os
import sys

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "3.2.0"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

from .libpytimemory import (
    report,
    toggle,
    enable,
    disable,
    is_enabled,
    enabled,
    has_mpi_support,
    set_rusage_children,
    set_rusage_self,
    timemory_init,
    timemory_finalize,
    initialize,
    finalize,
    timemory_argparse,
    add_arguments,
    get,
    get_text,
    clear,
    size,
    get_hash,
    get_hash_identifier,
    add_hash_id,
    mpi_init,
    mpi_finalize,
    upcxx_init,
    upcxx_finalize,
    dmp_init,
    dmp_finalize,
    start_mpip,
    stop_mpip,
    start_ompt,
    stop_ompt,
    start_ncclp,
    stop_ncclp,
    start_mallocp,
    stop_mallocp,
    enable_signal_detection,
    disable_signal_detection,
    set_exit_action,
    api,
    # papi,
    # cupti,
    # cuda,
    component,
    hardware_counters,
    manager,
    # profiler,
    # trace,
    units,
    scope,
    settings,
    storage,
    signals,
    region,
    options,
    timer_decorator,
    component_decorator,
    timer,
    rss_usage,
)

"""
from .libpytimemory import profiler as profile

__all__ = [
    "report",
    "toggle",
    "enable",
    "disable",
    "is_enabled",
    "enabled",
    "has_mpi_support",
    "set_rusage_children",
    "set_rusage_self",
    "timemory_init",
    "timemory_finalize",
    "initialize",
    "finalize",
    "timemory_argparse",
    "add_arguments",
    "get",
    "get_text",
    "clear",
    "size",
    "mpi_init",
    "mpi_finalize",
    "upcxx_init",
    "upcxx_finalize",
    "dmp_init",
    "dmp_finalize",
    "start_mpip",
    "stop_mpip",
    "start_ompt",
    "stop_ompt",
    "start_ncclp",
    "stop_ncclp",
    "start_mallocp",
    "stop_mallocp",
    "enable_signal_detection",
    "disable_signal_detection",
    "set_exit_action",
    "api",
    # papi,
    # cupti,
    # cuda,
    "component",
    "hardware_counters",
    "manager",
    "profiler",
    "trace",
    "units",
    "scope",
    "settings",
    "storage",
    "signals",
    "region",
    "options",
    "timer_decorator",
    "component_decorator",
    "timer",
    "rss_usage",
    "profile",
]
"""



# Tests of the C API

import unittest

import calipertest as cat

class CaliperCAPITest(unittest.TestCase):
    """ Caliper C API test cases """

    def test_c_ann_trace(self):
        target_cmd = [ './ci_test_c_ann', 'event-trace,output=stdout' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) >= 10)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, {'iteration', 'phase', 'time.inclusive.duration', 'global.int' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'event.end#phase': 'loop', 'phase': 'loop'}))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'event.end#iteration': '3', 'iteration': '3', 'phase': 'loop'}))
        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'attr.int', 'attr.dbl', 'attr.str', 'ci_test_c_ann.setbyname' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'attr.int' : '20', 'attr.str' : 'fidibus' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'test-attr-with-metadata' : 'abracadabra' }))

    def test_c_ann_globals(self):
        target_cmd = [ './ci_test_c_ann' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e', '--list-globals' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) == 1)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'global.double',
                         'global.string',
                         'global.int',
                         'global.uint',
                         'cali.caliper.version'
            }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'global.int'    : '1337',
                         'global.string' : 'my global string',
                         'global.uint'   : '42'
            }))

    def test_c_ann_metadata(self):
        target_cmd = [ './ci_test_c_ann' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e', '--list-attributes',
                       '--print-attributes', 'cali.attribute.name,cali.attribute.type,meta-attr' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'cali.attribute.name' : 'meta-attr',
                         'cali.attribute.type' : 'int' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'cali.attribute.name' : 'test-attr-with-metadata',
                         'cali.attribute.type' : 'string',
                         'meta-attr'           : '47' }))

    def test_c_ann_snapshot(self):
        target_cmd = [ './ci_test_c_snapshot' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) >= 4)
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'ci_test_c': 'snapshot', 'string_arg': 'teststring', 'int_arg': '42' }))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

from .metadatadb  import Attribute, Node, MetadataDB
from .readererror import ReaderError

class CaliperStreamReader:
    """ Reads a Caliper .cali data stream

    Use the read() method to read a Caliper .cali file and process
    records using callback methods.

    Attributes:
        globals     : dict
            The global key:value attributes defined in the .cali file
        num_records : dict
            The number of different kinds of records read
    """

    def __init__(self):
        self.db = MetadataDB()
        self.globals = {}
        self.num_records = {}


    def read(self, filename_or_stream, process_record_fn = None):
        """ Read a .cali file or stream.

        Snapshot records being read are fed to the user-provided
        process_record_fn callback function.
        After reading, global (metadata) values are stored in the
        globals attribute.

        Arguments:
            filename_or_stream : str or stream object
                The filename or stream to read.
            process_record_fn  : Function accepting a dict
                A callback function to process performance data records
        """

        if isinstance(filename_or_stream, str):
            with open(filename_or_stream) as f:
                for line in f:
                    self._process(line, process_record_fn)
        else:
            for line in filename_or_stream:
                self._process(line, process_record_fn)


    def attributes(self):
        """ Return the attribute keys.

        A file must have been read with read().

        Returns:
            The attribute keys (list of strings).
        """
        return self.db.attributes.keys()


    def attribute(self, attribute_name):
        """ Return a Caliper Attribute object for the given attribute name.
        """

        return self.db.attributes[attribute_name]


    def _process(self, line, process_record_fn = None):
        record = _read_cali_record(line)

        if not '__rec' in record:
            raise ReaderError('"__rec" missing: ' + line)

        kind = record['__rec'][0]

        if kind == 'node':
            self._process_node_record(record)
        elif kind == 'ctx' and process_record_fn is not None:
            process_record_fn(self._expand_record(record))
        elif kind == 'globals':
            self.globals = self._expand_record(record)

        if not kind in self.num_records:
            self.num_records[kind] = 0

        self.num_records[kind] += 1


    def _process_node_record(self, record):
        parent = Node.CALI_INV_ID

        if 'parent' in record:
            parent = int(record['parent'][0])

        self.db.import_node(int(record['id'][0]), int(record['attr'][0]), record['data'][0], parent)


    def _expand_record(self, record):
        result = {}

        if 'ref' in record:
            for node_id in record['ref']:
                for k, v in self._expand_node(int(node_id)).items():
                    if len(v) > 1:
                        result[k] = v
                    else:
                        result[k] = v[0]

        if 'attr' in record and 'data' in record:
            for attr_id, val in zip(record['attr'], record['data']):
                result[self.db.nodes[int(attr_id)].data] = val

        return result


    def _expand_node(self, node_id):
        def add(result, key, data):
            if key not in result:
                result[key] = []
            result[key].insert(0, data)

        result = {}
        node = self.db.nodes[node_id]

        while node is not None:
            attr = node.attribute()

            add(result, attr.name(), node.data)
            if attr.is_nested():
                add(result, 'path', node.data)

            node = node.parent

        return result


def _split_escape_string(input, keep_escape = True, splitchar = ',', escapechar = '\\'):
    """ Splits a string but allows escape characters """

    results = []
    string  = ""
    escaped = False

    for c in input:
        if not escaped and c == escapechar:
            escaped = True
            if keep_escape:
                string += c
        elif not escaped and c == splitchar:
            results.append(string)
            string = ""
        else:
            string += c
            escaped = False

    results.append(string)

    return results


def _read_cali_record(line):
    record  = {}
    entries = _split_escape_string(line.strip())

    for e in entries:
        kv = _split_escape_string(e, False, '=')

        if len(kv) < 1:
            raise ReaderError('Invalid record: {}'.format(line))

        record[kv[0]] = kv[1:]

    return record



#!@PYTHON_EXECUTABLE@

import sys
import timemory
from timemory.profiler import profile


def fib(n):
    return n if n < 2 else (fib(n - 1) + fib(n - 2))


@profile(["wall_clock", "peak_rss"])
def run_profile(nfib):
    return fib(nfib)


if __name__ == "__main__":
    nfib = int(sys.argv[1]) if len(sys.argv) > 1 else 23
    ans = run_profile(nfib)
    print("Answer = {}".format(ans))
    timemory.finalize()



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components


def generate_latex(component, indent):
    """
    This function generates a type trait label for C++
    """
    return "{}\item \lstinlinec{}{}{}".format(" " * indent, "{", component, "}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--verbose", help="Enable verbosity", default=0, type=int)
    parser.add_argument("-i", "--indent", help="Indentation", default=4, type=int)
    args = parser.parse_args()

    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))

    components.sort()
    outdata = "\\begin{itemize}\n"
    for component in components:
        outdata += "{}\n".format(generate_latex(component, args.indent))

    outdata += "\\end{itemize}"
    print("{}".format(outdata))



#!@PYTHON_EXECUTABLE@

import timemory
from timemory.util import auto_tuple
import sys


def fib(n):
    return n if n < 2 else (fib(n - 1) + fib(n - 2))


def get_tools(extra=[]):
    """
    Return the LIKWID components
    """
    components = ["likwid_marker", "likwid_nvmarker"] + extra
    return [getattr(timemory.component, x) for x in components]


@auto_tuple(get_tools(), add_args=True)
def run_fib(n):
    return fib(n)


if __name__ == "__main__":

    nfib = int(sys.argv[1]) if len(sys.argv) > 1 else 34

    with auto_tuple(get_tools(), "%s(%i)" % (sys.argv[0], nfib)):
        ans = fib(nfib)
        ans += run_fib(nfib + 1)

    print("fibonacci(34) = {}".format(ans))



#!/usr/bin/env python
# MIT License
#
# Copyright (c) 2020, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import platform
import subprocess as sp

__all__ = [
    "get_shared_lib_ext",
    "get_library_prefixes",
    "find_library_path",
    "get_linked_libraries",
    "echo_dart_measurement",
    "execute",
    "add_preload",
    "post_process",
]

is_darwin = False
is_linux = False
is_windows = False


if platform.system() == "Darwin":
    is_darwin = True
elif platform.system() == "Linux":
    is_linux = True
elif platform.system() == "Windows":
    is_windows = True


def get_shared_lib_ext():
    """
    Get shared library extension
    """
    if is_darwin:
        return ".dylib"
    elif is_linux:
        return ".so"
    elif is_windows:
        return ".dll"
    else:
        return ".so"


def get_library_prefixes():
    """
    Get library search paths
    """
    envvars = []
    if is_darwin:
        envvars += ["DYLD_LIBRARY_PATH"]
    if is_linux:
        envvars += ["LD_LIBRARY_PATH", "LIBRARY_PATH"]

    _prefixes = []
    for envvar in envvars:
        _var = os.environ.get(envvar)
        if _var is not None:
            _prefixes += _var.split(":")
    return _prefixes


def find_library_path(fname):
    _prefixes = get_library_prefixes()
    _lib_ext = get_shared_lib_ext()

    if _lib_ext not in fname:
        return None

    fname = fname.strip("@rpath/")
    if os.path.exists(fname):
        return os.path.realpath(fname)
    else:
        for _prefix in _prefixes:
            _outp = os.path.join(_prefix, fname)
            if os.path.exists(_outp):
                return os.path.realpath(_outp)
            _outp = os.path.join(_prefix, "lib" + fname)
            if os.path.exists(_outp):
                return os.path.realpath(_outp)

    print("Unable to find path to library '{}'".format(fname))
    return None


def get_linked_libraries(exes, libs=[]):
    """
    Get a list of libraries linked to the list of executables
    and (optional) find realpath to libraries
    """
    _linked_libs = []

    for exe in exes:
        output = None
        exe = os.path.realpath(exe)
        if not os.path.exists(exe):
            print("Executable '{}' does not exist".format(exe))
            continue

        if is_darwin:
            result = sp.run(["otool", "-L", "{}".format(exe)], stdout=sp.PIPE)
            output = result.stdout.decode("utf-8").split()
        elif is_linux:
            result = sp.run(["ldd", "{}".format(exe)], stdout=sp.PIPE)
            output = result.stdout.decode("utf-8").split()

        if output is None:
            print("No linked libraries found for '{}'!".format(exe))
            continue

        for out in output:
            _libpath = find_library_path(out)
            if _libpath is not None:
                _linked_libs += [_libpath]

    for _lib in libs:
        _libpath = find_library_path(_lib)
        if _libpath is not None:
            _linked_libs += [_libpath]

    return _linked_libs


def echo_dart_measurement(_name, _path, _type="jpeg"):
    from timemory.common import dart_measurement_file

    dart_measurement_file(_name, _path, _type)


def execute(cmd, outf, keep_going=True, timeout=5 * 60):
    """Execute a command"""
    from timemory.common import popen

    popen(cmd, outf, keep_going, timeout)


def add_preload(libs):
    preload_env = None
    if is_linux:
        preload_env = "LD_PRELOAD"
    elif is_darwin:
        preload_env = "DYLD_INSERT_LIBRARIES"
        os.environ["DYLD_FORCE_FLAT_NAMESPACE"] = "1"
    else:
        return

    if isinstance(libs, list) or isinstance(libs, tuple):
        preload = ":".join(libs)
    elif libs is not None:
        preload = libs
    else:
        return

    if preload_env is not None:
        current_preload = os.environ.get(preload_env)
        if current_preload is not None:
            os.environ[preload_env] = "{}:{}".format(current_preload, preload)
        else:
            os.environ[preload_env] = "{}".format(preload)


def post_process(
    exe,
    fname,
    image_type="jpeg",
    echo_dart=False,
    liblist=[],
    args=["--no_strip_temp", "--functions"],
    generate=["text", "cum", "dot"],
    dot_args=[],
):

    linked = get_linked_libraries(exe)
    libs = []
    if linked is not None:
        for _lib in linked:
            signature = "--add_lib={}".format(_lib)
            if signature not in libs:
                libs += [signature]

    if "text" in generate:
        execute(
            ["google-pprof", "--text"] + libs + args + [exe, fname],
            "{}.txt".format(fname),
        )

    if "cum" in generate:
        execute(
            ["google-pprof", "--text", "--cum"] + libs + args + [exe, fname],
            "{}.cum.txt".format(fname),
        )

    if "dot" in generate:
        oname = "{}.dot".format(fname)
        iname = "{}.{}".format(fname, image_type)
        execute(["google-pprof", "--dot"] + libs + args + [exe, fname], oname)
        execute(
            ["dot"]
            + dot_args
            + ["-T{}".format(image_type), oname, "-o", iname],
            None,
        )
        if echo_dart:
            echo_dart_measurement(
                os.path.basename(iname), os.path.realpath(iname), image_type
            )

    if "callgrind" in generate:
        execute(
            ["google-pprof", "--callgrind"] + libs + args + [exe, fname],
            "{}.callgrind".format(fname),
        )



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import unittest
import timemory as tim
from timemory.bundle import marker


# --------------------------- helper functions ----------------------------------------- #
# compute fibonacci
def fibonacci(n, with_arg=True):
    with marker(
        ["wall_clock", "peak_rss", "current_peak_rss"],
        "fib({})".format(n) if with_arg else "",
        mode="full" if not with_arg else "defer",
    ):
        return (
            n
            if n < 2
            else (fibonacci(n - 1, with_arg) + fibonacci(n - 2, with_arg))
        )


# --------------------------- Storage Tests set ---------------------------------------- #
# Storage tests class
class TimemoryStorageTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # set up environment variables
        tim.settings.verbose = 1
        tim.settings.debug = False
        tim.settings.tree_output = True
        tim.settings.text_output = True
        tim.settings.cout_output = False
        tim.settings.json_output = True
        tim.settings.flamegraph_output = False
        tim.settings.mpi_thread = False
        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False
        tim.settings.parse()

        # generate data
        fibonacci(5, False)
        # fibonacci(8)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    # ---------------------------------------------------------------------------------- #
    # test handling wall-clock data from storage
    def test_storage_tree_data(self):
        """storage_tree_output"""

        data = []
        data.append(
            "{}".format(tim.get(hierarchy=True, components=["wall_clock"]))
        )
        data.append(
            "{}".format(
                tim.storage.WallClockStorage.dumps(
                    tim.storage.WallClockStorage.dmp_get_tree()
                )
            )
        )

        self.assertEqual(data[0], data[1])

    # ---------------------------------------------------------------------------------- #
    # test handling wall-clock data from storage
    def test_storage_flat_data(self):
        """storage_flat_output"""

        data = []
        data.append(
            "{}".format(tim.get(hierarchy=False, components=["wall_clock"]))
        )
        data.append(
            "{}".format(
                tim.storage.WallClockStorage.dumps(
                    tim.storage.WallClockStorage.dmp_get()
                )
            )
        )

        self.assertEqual(data[0], data[1])

    # ---------------------------------------------------------------------------------- #
    # test metadata storage
    def test_metadata_storage(self):
        """metadata_storage"""

        existing = tim.manager.get_metadata()
        existing["info"]["FOO"] = "bar"
        existing["info"]["BAR"] = [1.0, 2.0, 3.0]
        tim.manager.add_metadata("FOO", "bar")
        tim.manager.add_metadata("BAR", [1.0, 2.0, 3.0])

        self.assertDictEqual(existing, tim.manager.get_metadata())
        print("timemory metadata info:")
        for key, item in tim.manager.get_metadata()["info"].items():
            print(f"    {key} : {item}")


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

import os
import unittest
import numpy as np
import timemory as tim
from timemory.bundle import marker

rank = 0
try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401

    rank = MPI.COMM_WORLD.Get_rank()
except ImportError:
    pass

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Muhammad Haseeb"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


# --------------------------- helper functions ----------------------------------------- #
# compute fibonacci
def fibonacci(n, with_arg=True):
    mode = "full" if not with_arg else "defer"
    with marker(
        ["wall_clock", "peak_rss", "current_peak_rss"],
        "fib({})".format(n) if with_arg else "",
        mode=mode,
    ):
        arr = np.ones([100, 100], dtype=float)  # noqa: F841
        return (
            n
            if n < 2
            else (fibonacci(n - 1, with_arg) + fibonacci(n - 2, with_arg))
        )


# --------------------------- Hatchet Tests set ---------------------------------------- #
# Hatchet tests class
class TimemoryHatchetTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # set up environment variables
        tim.settings.verbose = 1
        tim.settings.debug = False
        tim.settings.tree_output = True
        tim.settings.text_output = True
        tim.settings.cout_output = False
        tim.settings.json_output = False
        tim.settings.flamegraph_output = False
        tim.settings.mpi_thread = False
        tim.settings.dart_output = True
        tim.settings.dart_count = 1
        tim.settings.banner = False
        tim.settings.parse()

        # generate data
        fibonacci(15, False)
        fibonacci(8)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    # ---------------------------------------------------------------------------------- #
    # test handling wall-clock data from hatchet
    def test_hatchet_1D_data(self):
        """hatchet_1D_data"""

        try:
            import hatchet as ht
        except ImportError:
            return

        data = tim.get(hierarchy=True, components=["wall_clock"])
        gf = ht.GraphFrame.from_timemory(data)

        if rank == 0:
            print(gf.dataframe)
            print(gf.tree("sum"))
            print(gf.tree("sum.inc"))

    # ---------------------------------------------------------------------------------- #
    # test handling multi-dimensional data from hatchet
    def test_hatchet_2D_data(self):
        """hatchet_2d_data"""

        try:
            import hatchet as ht
        except ImportError:
            return

        from timemory.component import CurrentPeakRss

        data = tim.get(hierarchy=True, components=[CurrentPeakRss.id()])
        gf = ht.GraphFrame.from_timemory(data)

        if rank == 0:
            print(gf.dataframe)
            print(gf.tree("sum.start-peak-rss"))
            print(gf.tree("sum.stop-peak-rss.inc"))

        self.assertTrue(gf.tree("sum.start-peak-rss.inc") is not None)
        self.assertTrue(gf.tree("sum.stop-peak-rss") is not None)

    def test_hatchet_analyze(self):
        """test_hatchet_analyze"""

        try:
            import hatchet as ht  # noqa: F401
        except ImportError:
            return

        from timemory.analyze import embedded_analyze
        from timemory import settings
        import timemory

        outpath = os.path.join(
            settings.output_path,
            "analysis",
        )

        args = ["--expression", "x > 0"] + (
            f"-f dot flamegraph tree -o {outpath} --per-thread --per-rank "
            + "--select peak_rss --search fibonacci -e"
        ).split()

        if rank == 0:
            print(f"arguments: {args}")

        embedded_analyze(
            args,
            data=[timemory.get(hierarchy=True)],
        )


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



# IoService tests

import json
import unittest

import calipertest as cat

class CaliperIoServiceTest(unittest.TestCase):
    """ Caliper I/O service test case """

    def test_ioservice(self):
        target_cmd = [ './ci_test_io', 'event-trace(trace.io=true),output=stdout' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'io.bytes.read': '16',
                         'io.region':     'read',
                         'function':      'main' }))

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'io.bytes.read',
                         'io.region',
                         'io.filesystem',
                         'io.mount.point' }))

    def test_ioservice_multichannel_disabled(self):
        # Enable the IO service in the default config but not in the trace channel.
        # While I/O regions will be on the shared blackboard, there should be no
        # I/O event records in the trace channel.

        target_cmd = [ './ci_test_io', 'event-trace(trace.io=false,output=stdout)' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0',
            'CALI_SERVICES_ENABLE'   : 'event,io,trace',
            'CALI_CHANNEL_CONFIG_CHECK' : 'false'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'function': 'main' }))
        self.assertFalse(cat.has_snapshot_with_keys(
            snapshots, { 'io.region' }))

    def test_hatchet_profile_ioservice(self):
        # Test the I/O bytes metric in the
        target_cmd = [ './ci_test_io', 'hatchet-region-profile,use.mpi=false,output=stdout,output.format=json,io.bytes.read=true' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test(target_cmd, caliper_config)[0] )

        self.assertEqual(obj[1]['path'], 'main')
        self.assertEqual(int(obj[1]['Bytes read']), 16)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import hatchet as ht


if __name__ == "__main__":
    # Path to DOT file.
    dot_file = "../../../hatchet/tests/data/gprof2dot-cpi/callgrind.dot.64042.0.1"

    # Use hatchet's ``from_gprof_dot`` API to read in the DOT file. The result
    # is stored into Hatchet's GraphFrame.
    gf = ht.GraphFrame.from_gprof_dot(dot_file)

    # Printout the DataFrame component of the GraphFrame.
    print(gf.dataframe)

    # Printout the graph component of the GraphFrame.
    # Because no metric parameter is specified, ``time`` is used by default.
    print(gf.tree())



#!/usr/bin/env python
# MIT License
#
# Copyright (c) 2020, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from . import utils


__all__ = ["execute"]


def execute(
    cmd,
    prefix="cpu.prof",
    freq=250,
    malloc_stats=0,
    realtime=1,
    preload=True,
    selected=0,
    image_type="jpeg",
    echo_dart=False,
    libs=[],
    args=["--no_strip_temp", "--functions"],
    generate=["text", "cum", "dot"],
    dot_args=[],
):

    os.environ["CPUPROFILE_FREQUENCY"] = "{}".format(freq)
    os.environ["MALLOCSTATS"] = "{}".format(malloc_stats)
    os.environ["CPUPROFILE_REALTIME"] = "{}".format(realtime)
    os.environ["PROFILESELECTED"] = "{}".format(selected)

    if preload:
        _libpath = None
        for libname in ("profiler", "tcmalloc_and_profiler"):
            if _libpath is None:
                _libpath = utils.find_library_path(libname)
        if _libpath is None:
            raise RuntimeError("Preload failed. Cannot find libprofiler")
        utils.add_preload(_libpath)

    exe = os.path.basename(cmd[0])
    prefix += ".{}".format(exe)
    if not os.path.exists(prefix):
        os.makedirs(prefix)
    N = 0
    fprefix = os.path.join(prefix, "gperf.")
    fname = fprefix + "{}".format(N)
    while os.path.exists(fname):
        N += 1
        fname = fprefix + "{}".format(N)

    os.environ["CPUPROFILE"] = "{}".format(fname)

    print("exeuting: {}, exe: {}".format(cmd, exe))
    utils.execute(cmd, "{}.log".format(fname))

    utils.post_process(
        cmd[0], fname, image_type, echo_dart, libs, args, generate, dot_args
    )



#!/usr/bin/env python

import argparse
import hatchet as ht

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m", "--metric", default="sum", type=str, help="Metric to display"
    )
    args, argv = parser.parse_known_args()
    files = argv[:]
    if not files:
        files.append("../../../hatchet/tests/data/timemory/wall.tree.json")

    for json_file in files:
        # Use hatchet's ``from_timemory`` API with the hierarchical json output.
        # The result is stored into Hatchet's GraphFrame.
        gf = ht.GraphFrame.from_timemory(json_file)

        # Printout the DataFrame component of the GraphFrame.
        print(gf.dataframe)

        # Printout the graph component of the GraphFrame.
        # one dimensional components use standardize metric labels,
        # e.g., sum, sum.inc, mean, mean.inc, etc.
        # but multi-dimensional data, such as the current-peak-rss
        # (which reports the peak RSS at the start of the marker
        # and at the end) and hardware-counter data, create
        # unique suffixes, e.g. sum.start-peak-rss,
        # sum.start-peak-rss, et.c
        print(gf.tree(args.metric))



# Loop report tests

import json
import os
import subprocess
import unittest

import calipertest as cat


class CaliperSpotControllerTest(unittest.TestCase):
    def test_spot_regionprofile(self):
        target_cmd = [ './ci_test_macros', '0', 'spot(output=stdout)' ]
        query_cmd = [ '../../src/tools/cali-query/cali-query', '-q', 'format json(object)' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config) )

        self.assertIn('globals', obj)
        self.assertIn('records', obj)

        self.assertIn('spot.format.version', obj['globals'])
        self.assertIn('avg#inclusive#sum#time.duration', obj['globals']['spot.metrics'])
        self.assertIn('min#inclusive#sum#time.duration', obj['globals']['spot.metrics'])
        self.assertIn('max#inclusive#sum#time.duration', obj['globals']['spot.metrics'])
        self.assertIn('sum#inclusive#sum#time.duration', obj['globals']['spot.metrics'])
        self.assertIn('avg#inclusive#sum#time.duration', obj['attributes'])
        self.assertIn('min#inclusive#sum#time.duration', obj['attributes'])
        self.assertIn('max#inclusive#sum#time.duration', obj['attributes'])
        self.assertIn('sum#inclusive#sum#time.duration', obj['attributes'])
        self.assertIn('attribute.alias', obj['attributes']['avg#inclusive#sum#time.duration'])
        self.assertIn('attribute.unit',  obj['attributes']['avg#inclusive#sum#time.duration'])

        self.assertEqual('Avg time/rank', obj['attributes']['avg#inclusive#sum#time.duration']['attribute.alias'])
        self.assertEqual('sec',           obj['attributes']['avg#inclusive#sum#time.duration']['attribute.unit'])

        r = None
        for rec in obj['records']:
            if 'path' in rec and rec['path'] == 'main':
                r = rec
                break

        self.assertTrue(r is not None, 'No record with "path" entry found')
        self.assertIn('avg#inclusive#sum#time.duration', r)
        self.assertIn('min#inclusive#sum#time.duration', r)
        self.assertIn('max#inclusive#sum#time.duration', r)
        self.assertIn('sum#inclusive#sum#time.duration', r)
        self.assertEqual(r['spot.channel'], 'regionprofile')

    def test_spot_timeseries(self):
        target_cmd = [ './ci_test_macros', '0', 'spot(output=stdout,timeseries,timeseries.iteration_interval=15)', '75' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-q', 'select * where spot.channel=timeseries format json(object)' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config) )

        self.assertIn('globals', obj)
        self.assertIn('records', obj)

        self.assertIn('timeseries', obj['globals']['spot.options'])

        self.assertIn('avg#loop.iterations/time.duration', obj['globals']['spot.timeseries.metrics'])
        self.assertIn('avg#loop.iterations/time.duration', obj['records'][0])
        self.assertEqual(obj['records'][0]['spot.channel'], 'timeseries')

        blocks = []
        for rec in obj['records']:
            blocks.append(int(rec['block']))
        blocks.sort()

        self.assertEqual(blocks, [ 0, 15, 30, 45, 60 ])

    def test_spot_optioncheck_a(self):
        target_cmd = [ './ci_test_macros', '0', 'spot(output=stdout,timeseries.iteration_interval=15)' ]

        env = {
            'CALI_LOG_VERBOSITY' : '0',
        }

        log_targets = [
            'timeseries.iteration_interval is set but the timeseries option is not enabled'
        ]

        target_proc = subprocess.Popen(target_cmd, env=env, stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        _,report_err = target_proc.communicate()

        lines = report_err.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_spot_optioncheck_b(self):
        target_cmd = [ './ci_test_macros', '0', 'spot(output=stdout,timeseries,timeseries.metrics=blagarbl)' ]

        env = {
            'CALI_LOG_VERBOSITY' : '0',
        }

        log_targets = [
            'Unknown option: blagarbl'
        ]

        target_proc = subprocess.Popen(target_cmd, env=env, stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        _,report_err = target_proc.communicate()

        lines = report_err.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

if __name__ == "__main__":
    unittest.main()



# CpuInfo tests

import unittest

import calipertest as calitest

class CaliperCpuInfoTest(unittest.TestCase):
    """ Caliper CpuInfo test case """

    def test_papi(self):
        target_cmd = [ './ci_test_basic' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'event,cpuinfo,trace,recorder',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = calitest.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = calitest.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(calitest.has_snapshot_with_keys(
            snapshots, { 'cpuinfo.cpu',
                         'cpuinfo.numa_node',
                         'phase',
                         'iteration' }))


if __name__ == "__main__":
    unittest.main()



# Logging tests

import unittest

import calipertest as cat


class CaliperLogTest(unittest.TestCase):
    """ Caliper Log test cases """

    def test_log_verbose(self):
        target_cmd = [ './ci_test_basic' ]

        env = { 'CALI_LOG_VERBOSITY' : '3',
                'CALI_LOG_LOGFILE'   : 'stdout'
        }

        log_targets = [
            'CALI_LOG_VERBOSITY=3',
            '== CALIPER: default: snapshot scopes: process thread',
            '== CALIPER: Releasing channel default',
            '== CALIPER: Releasing Caliper thread data',
            'Process blackboard',
            'Thread blackboard',
            'Metadata tree',
            'Metadata memory pool'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_cali_config_error_msg(self):
        target_cmd = [ './ci_test_basic' ]

        env = {
            'CALI_LOG_VERBOSITY' : '0',
            'CALI_LOG_LOGFILE'   : 'stdout',
            'CALI_CONFIG'        : 'blagarbl'
        }

        log_targets = [
            '== CALIPER: CALI_CONFIG: error: Unknown config or parameter: blagarbl'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_scope_parse_error_msgs(self):
        target_cmd = [ './ci_test_basic' ]

        env = {
            'CALI_LOG_VERBOSITY' : '0',
            'CALI_LOG_LOGFILE'   : 'stdout',
            'CALI_CHANNEL_SNAPSHOT_SCOPES' : 'foo',
            'CALI_CALIPER_ATTRIBUTE_DEFAULT_SCOPE' : 'bar'
        }

        log_targets = [
            'Invalid value "foo" for CALI_CHANNEL_SNAPSHOT_SCOPES',
            'Invalid value "bar" for CALI_CALIPER_ATTRIBUTE_DEFAULT_SCOPE'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_log_silent(self):
        target_cmd = [ './ci_test_basic' ]

        env = { 'CALI_LOG_VERBOSITY' : '0',
                'CALI_LOG_LOGFILE'   : 'stdout'
        }

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        self.assertTrue(len(lines) == 0)

if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import unittest

import caliperreader as cr

class CaliperReaderBasic(unittest.TestCase):
    """ Basic test cases for the Caliper reader """

    def test_reader_class(self):
        r = cr.CaliperReader()

        r.read('example-profile.cali')

        self.assertEqual(len(r.records), 25)
        self.assertEqual(len(r.globals), 22)

        self.assertEqual(r.globals['spot.format.version'], '2')
        self.assertEqual(r.globals['threads'], '2')

        self.assertIn('path', r.records[4])
        self.assertIn('function', r.records[5])
        self.assertIn('avg#inclusive#sum#time.duration', r.records[6])

        self.assertTrue( r.attribute('function').is_nested())
        self.assertFalse(r.attribute('function').is_value() )

        self.assertFalse(r.attribute('avg#inclusive#sum#time.duration').is_nested())
        self.assertTrue( r.attribute('avg#inclusive#sum#time.duration').is_value() )

        self.assertEqual(r.attribute('function').attribute_type(), 'string')
        self.assertEqual(r.attribute('figure_of_merit').get('adiak.type'), 'double')

        self.assertIsNotNone(r.attribute('avg#inclusive#sum#time.duration').get('attribute.unit'))
        self.assertIsNone(r.attribute('function').get('attribute.unit'))
        self.assertIsNone(r.attribute('function').get('DOES NOT EXIST'))

    def test_read_contents(self):
        records, globals = cr.read_caliper_contents('example-profile.cali')

        self.assertEqual(len(records), 25)
        self.assertEqual(len(globals), 22)

        self.assertEqual(globals['cali.channel'],   'spot')
        self.assertEqual(globals['mpi.world.size'], '1')
        self.assertEqual(globals['problem_size'],   '30')

        self.assertIn('path',     records[7])
        self.assertIn('function', records[8])
        self.assertIn('avg#inclusive#sum#time.duration', records[9])

    def test_read_globals(self):
        globals = cr.read_caliper_globals('example-profile.cali')

        self.assertEqual(len(globals), 22)

        self.assertIn('libraries', globals)
        self.assertIn('elapsed_time', globals)
        self.assertIn('cali.caliper.version', globals)

        self.assertEqual(globals['region_balance'], '1')
        self.assertEqual(globals['user'], 'david')

if __name__ == "__main__":
    unittest.main()



#!@PYTHON_EXECUTABLE@

import timemory
from timemory.bundle import auto_tuple
from timemory.component import CaliperConfig
import sys


def fib(n):
    return n if n < 2 else (fib(n - 1) + fib(n - 2))


def get_tools(extra=[]):
    """
    Return the caliper component
    """
    components = ["caliper_marker"] + extra
    return [getattr(timemory.component, x) for x in components]


@auto_tuple(get_tools(), add_args=True)
def run_fib(n):
    return fib(n)


if __name__ == "__main__":

    nfib = int(sys.argv[1]) if len(sys.argv) > 1 else 34

    cfg = CaliperConfig()
    cfg.configure("runtime-report")
    cfg.start()

    with auto_tuple(get_tools(), "%s(%i)" % (sys.argv[0], nfib)):
        ans = fib(nfib)
        ans += run_fib(nfib + 1)

    cfg.stop()
    print("fibonacci({}) = {}".format(nfib, ans))
    timemory.finalize()
